# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
from typing import List, Dict, Optional
from pyomo.common.collections import ComponentMap
from pyomo.common.dependencies import attempt_import
from pyomo.common.errors import PyomoException
from pyomo.common.flags import NOTSET
from pyomo.common.timing import HierarchicalTimer
from pyomo.common.config import ConfigValue, NonNegativeInt
from pyomo.common.tee import TeeStream, capture_output
from pyomo.common.log import LogStream
from pyomo.core.kernel.objective import minimize, maximize
from pyomo.core.base import SymbolMap
from pyomo.core.base.var import VarData
from pyomo.core.base.constraint import ConstraintData
from pyomo.core.base.sos import SOSConstraintData
from pyomo.core.base.param import ParamData
from pyomo.core.expr.numvalue import value, is_constant
from pyomo.repn import generate_standard_repn
from pyomo.core.expr.numeric_expr import NPV_MaxExpression, NPV_MinExpression
from pyomo.contrib.appsi.base import (
    PersistentSolver,
    Results,
    TerminationCondition,
    MIPSolverConfig,
    PersistentBase,
    PersistentSolutionLoader,
)
from pyomo.contrib.appsi.cmodel import cmodel, cmodel_available
from pyomo.common.dependencies import numpy as np
from pyomo.core.staleflag import StaleFlagManager
import sys

logger = logging.getLogger(__name__)

highspy, highspy_available = attempt_import('highspy')


class DegreeError(PyomoException):
    pass


class HighsConfig(MIPSolverConfig):
    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        super(HighsConfig, self).__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )

        self.declare('logfile', ConfigValue(domain=str))
        self.declare('solver_output_logger', ConfigValue())
        self.declare('log_level', ConfigValue(domain=NonNegativeInt))

        self.logfile = ''
        self.solver_output_logger = logger
        self.log_level = logging.INFO


class HighsResults(Results):
    def __init__(self, solver):
        super().__init__()
        self.wallclock_time = None
        self.solution_loader = PersistentSolutionLoader(solver=solver)


class _MutableVarBounds:
    def __init__(self, lower_expr, upper_expr, pyomo_var_id, var_map, highs):
        self.pyomo_var_id = pyomo_var_id
        self.lower_expr = lower_expr
        self.upper_expr = upper_expr
        self.var_map = var_map
        self.highs = highs

    def update(self):
        col_ndx = self.var_map[self.pyomo_var_id]
        lb = value(self.lower_expr)
        ub = value(self.upper_expr)
        self.highs.changeColBounds(col_ndx, lb, ub)


class _MutableLinearCoefficient:
    def __init__(self, pyomo_con, pyomo_var_id, con_map, var_map, expr, highs):
        self.expr = expr
        self.highs = highs
        self.pyomo_var_id = pyomo_var_id
        self.pyomo_con = pyomo_con
        self.con_map = con_map
        self.var_map = var_map

    def update(self):
        row_ndx = self.con_map[self.pyomo_con]
        col_ndx = self.var_map[self.pyomo_var_id]
        self.highs.changeCoeff(row_ndx, col_ndx, value(self.expr))


class _MutableObjectiveCoefficient:
    def __init__(self, pyomo_var_id, var_map, expr, highs):
        self.expr = expr
        self.highs = highs
        self.pyomo_var_id = pyomo_var_id
        self.var_map = var_map

    def update(self):
        col_ndx = self.var_map[self.pyomo_var_id]
        self.highs.changeColCost(col_ndx, value(self.expr))


class _MutableObjectiveOffset:
    def __init__(self, expr, highs):
        self.expr = expr
        self.highs = highs

    def update(self):
        self.highs.changeObjectiveOffset(value(self.expr))


class _MutableConstraintBounds:
    def __init__(self, lower_expr, upper_expr, pyomo_con, con_map, highs):
        self.lower_expr = lower_expr
        self.upper_expr = upper_expr
        self.con = pyomo_con
        self.con_map = con_map
        self.highs = highs

    def update(self):
        row_ndx = self.con_map[self.con]
        lb = value(self.lower_expr)
        ub = value(self.upper_expr)
        self.highs.changeRowBounds(row_ndx, lb, ub)


class Highs(PersistentBase, PersistentSolver):
    """
    Interface to HiGHS
    """

    _available = None

    def __init__(self, only_child_vars=False):
        super().__init__(only_child_vars=only_child_vars)
        self._config = HighsConfig()
        self._solver_options = dict()
        self._solver_model = None
        self._pyomo_var_to_solver_var_map = dict()
        self._pyomo_con_to_solver_con_map = dict()
        self._solver_con_to_pyomo_con_map = dict()
        self._mutable_helpers = dict()
        self._mutable_bounds = dict()
        self._objective_helpers = list()
        self._last_results_object: Optional[HighsResults] = None
        self._sol = None

    def available(self):
        if highspy_available:
            return self.Availability.FullLicense
        else:
            return self.Availability.NotFound

    def version(self):
        try:
            version = (
                highspy.HIGHS_VERSION_MAJOR,
                highspy.HIGHS_VERSION_MINOR,
                highspy.HIGHS_VERSION_PATCH,
            )
        except AttributeError:
            # Older versions of Highs do not have the above attributes
            # and the solver version can only be obtained by making
            # an instance of the solver class.
            tmp = highspy.Highs()
            version = (tmp.versionMajor(), tmp.versionMinor(), tmp.versionPatch())

        return version

    @property
    def config(self) -> HighsConfig:
        return self._config

    @config.setter
    def config(self, val: HighsConfig):
        self._config = val

    @property
    def highs_options(self):
        """
        A dictionary mapping solver options to values for those options. These
        are solver specific.

        Returns
        -------
        dict
           A dictionary mapping solver options to values for those options
        """
        return self._solver_options

    @highs_options.setter
    def highs_options(self, val: Dict):
        self._solver_options = val

    @property
    def symbol_map(self):
        return SymbolMap()
        # raise RuntimeError('Highs interface does not have a symbol map')

    def warm_start_capable(self):
        return True

    def _warm_start(self):
        # Collect all variable values
        col_value = np.zeros(len(self._pyomo_var_to_solver_var_map))
        has_values = False

        for var_id, col_ndx in self._pyomo_var_to_solver_var_map.items():
            var = self._vars[var_id][0]
            if var.value is not None:
                col_value[col_ndx] = value(var)
                has_values = True

        if has_values:
            solution = highspy.HighsSolution()
            solution.col_value = col_value
            solution.value_valid = True
            solution.dual_valid = False

            # Set the solution as a MIP start
            self._solver_model.setSolution(solution)

    def _solve(self, timer: HierarchicalTimer):
        config = self.config
        options = self.highs_options

        ostreams = [
            LogStream(
                level=self.config.log_level, logger=self.config.solver_output_logger
            )
        ]
        if self.config.stream_solver:
            ostreams.append(sys.stdout)

        with capture_output(output=TeeStream(*ostreams), capture_fd=True):
            self._solver_model.setOptionValue('log_to_console', True)
            if config.logfile != '':
                self._solver_model.setOptionValue('log_file', config.logfile)

            if config.time_limit is not None:
                self._solver_model.setOptionValue('time_limit', config.time_limit)
            if config.mip_gap is not None:
                self._solver_model.setOptionValue('mip_rel_gap', config.mip_gap)

            for key, option in options.items():
                self._solver_model.setOptionValue(key, option)

            if config.warmstart:
                self._warm_start()
            timer.start('optimize')
            if self.version()[:2] >= (1, 8):
                self._solver_model.HandleKeyboardInterrupt = True
            self._solver_model.run()
            timer.stop('optimize')

        return self._postsolve(timer)

    def solve(self, model, timer: HierarchicalTimer = None) -> Results:
        StaleFlagManager.mark_all_as_stale()
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        if timer is None:
            timer = HierarchicalTimer()
        if model is not self._model:
            timer.start('set_instance')
            self.set_instance(model)
            timer.stop('set_instance')
        else:
            timer.start('update')
            self.update(timer=timer)
            timer.stop('update')
        res = self._solve(timer)
        self._last_results_object = res
        if self.config.report_timing:
            logger.info('\n' + str(timer))
        return res

    def _process_domain_and_bounds(self, var_id):
        _v, _lb, _ub, _fixed, _domain_interval, _value = self._vars[var_id]
        lb, ub, step = _domain_interval
        if lb is None:
            lb = -highspy.kHighsInf
        if ub is None:
            ub = highspy.kHighsInf
        if step == 0:
            vtype = highspy.HighsVarType.kContinuous
        elif step == 1:
            vtype = highspy.HighsVarType.kInteger
        else:
            raise ValueError(
                f'Unrecognized domain step: {step} (should be either 0 or 1)'
            )
        if _fixed:
            lb = _value
            ub = _value
        else:
            if _lb is not None or _ub is not None:
                if not is_constant(_lb) or not is_constant(_ub):
                    if _lb is None:
                        tmp_lb = -highspy.kHighsInf
                    else:
                        tmp_lb = _lb
                    if _ub is None:
                        tmp_ub = highspy.kHighsInf
                    else:
                        tmp_ub = _ub
                    mutable_bound = _MutableVarBounds(
                        lower_expr=NPV_MaxExpression((tmp_lb, lb)),
                        upper_expr=NPV_MinExpression((tmp_ub, ub)),
                        pyomo_var_id=var_id,
                        var_map=self._pyomo_var_to_solver_var_map,
                        highs=self._solver_model,
                    )
                    self._mutable_bounds[var_id] = (_v, mutable_bound)
            if _lb is not None:
                lb = max(value(_lb), lb)
            if _ub is not None:
                ub = min(value(_ub), ub)

        return lb, ub, vtype

    def _add_variables(self, variables: List[VarData]):
        self._sol = None
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        lbs = list()
        ubs = list()
        indices = list()
        vtypes = list()

        current_num_vars = len(self._pyomo_var_to_solver_var_map)
        for v in variables:
            v_id = id(v)
            lb, ub, vtype = self._process_domain_and_bounds(v_id)
            lbs.append(lb)
            ubs.append(ub)
            vtypes.append(vtype)
            indices.append(current_num_vars)
            self._pyomo_var_to_solver_var_map[v_id] = current_num_vars
            current_num_vars += 1

        self._solver_model.addVars(
            len(lbs), np.array(lbs, dtype=np.double), np.array(ubs, dtype=np.double)
        )
        self._solver_model.changeColsIntegrality(
            len(vtypes), np.array(indices), np.array(vtypes)
        )

    def _add_params(self, params: List[ParamData]):
        pass

    def _reinit(self):
        saved_config = self.config
        saved_options = self.highs_options
        saved_update_config = self.update_config
        self.__init__(only_child_vars=self._only_child_vars)
        self.config = saved_config
        self.highs_options = saved_options
        self.update_config = saved_update_config

    def set_instance(self, model):
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        if not self.available():
            c = self.__class__
            raise PyomoException(
                f'Solver {c.__module__}.{c.__qualname__} is not available '
                f'({self.available()}).'
            )

        ostreams = [
            LogStream(
                level=self.config.log_level, logger=self.config.solver_output_logger
            )
        ]
        if self.config.stream_solver:
            ostreams.append(sys.stdout)
        with capture_output(output=TeeStream(*ostreams), capture_fd=True):
            self._reinit()
            self._model = model
            if self.use_extensions and cmodel_available:
                self._expr_types = cmodel.PyomoExprTypes()

            self._solver_model = highspy.Highs()
            self.add_block(model)
            if self._objective is None:
                self.set_objective(None)

    def _add_constraints(self, cons: List[ConstraintData]):
        self._sol = None
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        current_num_cons = len(self._pyomo_con_to_solver_con_map)
        lbs = list()
        ubs = list()
        starts = list()
        var_indices = list()
        coef_values = list()

        for con in cons:
            repn = generate_standard_repn(
                con.body, quadratic=False, compute_values=False
            )
            if repn.nonlinear_expr is not None:
                raise DegreeError(
                    f'Highs interface does not support expressions of degree {repn.polynomial_degree()}'
                )

            starts.append(len(coef_values))
            for ndx, coef in enumerate(repn.linear_coefs):
                v = repn.linear_vars[ndx]
                v_id = id(v)
                coef_val = value(coef)
                if not is_constant(coef):
                    mutable_linear_coefficient = _MutableLinearCoefficient(
                        pyomo_con=con,
                        pyomo_var_id=v_id,
                        con_map=self._pyomo_con_to_solver_con_map,
                        var_map=self._pyomo_var_to_solver_var_map,
                        expr=coef,
                        highs=self._solver_model,
                    )
                    if con not in self._mutable_helpers:
                        self._mutable_helpers[con] = list()
                    self._mutable_helpers[con].append(mutable_linear_coefficient)
                    if coef_val == 0:
                        continue
                var_indices.append(self._pyomo_var_to_solver_var_map[v_id])
                coef_values.append(coef_val)

            if con.has_lb():
                lb = con.lower - repn.constant
            else:
                lb = -highspy.kHighsInf
            if con.has_ub():
                ub = con.upper - repn.constant
            else:
                ub = highspy.kHighsInf

            if not is_constant(lb) or not is_constant(ub):
                mutable_con_bounds = _MutableConstraintBounds(
                    lower_expr=lb,
                    upper_expr=ub,
                    pyomo_con=con,
                    con_map=self._pyomo_con_to_solver_con_map,
                    highs=self._solver_model,
                )
                if con not in self._mutable_helpers:
                    self._mutable_helpers[con] = [mutable_con_bounds]
                else:
                    self._mutable_helpers[con].append(mutable_con_bounds)

            lbs.append(value(lb))
            ubs.append(value(ub))
            self._pyomo_con_to_solver_con_map[con] = current_num_cons
            self._solver_con_to_pyomo_con_map[current_num_cons] = con
            current_num_cons += 1

        self._solver_model.addRows(
            len(lbs),
            np.array(lbs, dtype=np.double),
            np.array(ubs, dtype=np.double),
            len(coef_values),
            np.array(starts),
            np.array(var_indices),
            np.array(coef_values, dtype=np.double),
        )

    def _add_sos_constraints(self, cons: List[SOSConstraintData]):
        if cons:
            raise NotImplementedError(
                'Highs interface does not support SOS constraints'
            )

    def _remove_constraints(self, cons: List[ConstraintData]):
        self._sol = None
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        indices_to_remove = list()
        for con in cons:
            con_ndx = self._pyomo_con_to_solver_con_map.pop(con)
            del self._solver_con_to_pyomo_con_map[con_ndx]
            indices_to_remove.append(con_ndx)
            self._mutable_helpers.pop(con, None)
        self._solver_model.deleteRows(
            len(indices_to_remove), np.sort(np.array(indices_to_remove))
        )
        con_ndx = 0
        new_con_map = dict()
        for c in self._pyomo_con_to_solver_con_map.keys():
            new_con_map[c] = con_ndx
            con_ndx += 1
        self._pyomo_con_to_solver_con_map.clear()
        self._pyomo_con_to_solver_con_map.update(new_con_map)
        self._solver_con_to_pyomo_con_map.clear()
        self._solver_con_to_pyomo_con_map.update(
            {v: k for k, v in self._pyomo_con_to_solver_con_map.items()}
        )

    def _remove_sos_constraints(self, cons: List[SOSConstraintData]):
        if cons:
            raise NotImplementedError(
                'Highs interface does not support SOS constraints'
            )

    def _remove_variables(self, variables: List[VarData]):
        self._sol = None
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        indices_to_remove = list()
        for v in variables:
            v_id = id(v)
            v_ndx = self._pyomo_var_to_solver_var_map.pop(v_id)
            indices_to_remove.append(v_ndx)
            self._mutable_bounds.pop(v_id, None)
        indices_to_remove.sort()
        self._solver_model.deleteVars(
            len(indices_to_remove), np.array(indices_to_remove)
        )
        v_ndx = 0
        new_var_map = dict()
        for v_id in self._pyomo_var_to_solver_var_map.keys():
            new_var_map[v_id] = v_ndx
            v_ndx += 1
        self._pyomo_var_to_solver_var_map.clear()
        self._pyomo_var_to_solver_var_map.update(new_var_map)

    def _remove_params(self, params: List[ParamData]):
        pass

    def _update_variables(self, variables: List[VarData]):
        self._sol = None
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        indices = list()
        lbs = list()
        ubs = list()
        vtypes = list()

        for v in variables:
            v_id = id(v)
            self._mutable_bounds.pop(v_id, None)
            v_ndx = self._pyomo_var_to_solver_var_map[v_id]
            lb, ub, vtype = self._process_domain_and_bounds(v_id)
            lbs.append(lb)
            ubs.append(ub)
            vtypes.append(vtype)
            indices.append(v_ndx)

        self._solver_model.changeColsBounds(
            len(indices),
            np.array(indices),
            np.array(lbs, dtype=np.double),
            np.array(ubs, dtype=np.double),
        )
        self._solver_model.changeColsIntegrality(
            len(indices), np.array(indices), np.array(vtypes)
        )

    def update_params(self):
        self._sol = None
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        for con, helpers in self._mutable_helpers.items():
            for helper in helpers:
                helper.update()
        for k, (v, helper) in self._mutable_bounds.items():
            helper.update()
        for helper in self._objective_helpers:
            helper.update()

    def _set_objective(self, obj):
        self._sol = None
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        n = len(self._pyomo_var_to_solver_var_map)
        indices = np.arange(n)
        costs = np.zeros(n, dtype=np.double)
        self._objective_helpers = list()
        if obj is None:
            sense = highspy.ObjSense.kMinimize
            self._solver_model.changeObjectiveOffset(0)
        else:
            if obj.sense == minimize:
                sense = highspy.ObjSense.kMinimize
            elif obj.sense == maximize:
                sense = highspy.ObjSense.kMaximize
            else:
                raise ValueError(
                    'Objective sense is not recognized: {0}'.format(obj.sense)
                )

            repn = generate_standard_repn(
                obj.expr, quadratic=False, compute_values=False
            )
            if repn.nonlinear_expr is not None:
                raise DegreeError(
                    f'Highs interface does not support expressions of degree {repn.polynomial_degree()}'
                )

            for coef, v in zip(repn.linear_coefs, repn.linear_vars):
                v_id = id(v)
                v_ndx = self._pyomo_var_to_solver_var_map[v_id]
                costs[v_ndx] = value(coef)
                if not is_constant(coef):
                    mutable_objective_coef = _MutableObjectiveCoefficient(
                        pyomo_var_id=v_id,
                        var_map=self._pyomo_var_to_solver_var_map,
                        expr=coef,
                        highs=self._solver_model,
                    )
                    self._objective_helpers.append(mutable_objective_coef)

            self._solver_model.changeObjectiveOffset(value(repn.constant))
            if not is_constant(repn.constant):
                mutable_objective_offset = _MutableObjectiveOffset(
                    expr=repn.constant, highs=self._solver_model
                )
                self._objective_helpers.append(mutable_objective_offset)

        self._solver_model.changeObjectiveSense(sense)
        self._solver_model.changeColsCost(n, indices, costs)

    def _postsolve(self, timer: HierarchicalTimer):
        config = self.config

        highs = self._solver_model
        status = highs.getModelStatus()

        results = HighsResults(self)
        results.wallclock_time = highs.getRunTime()

        if status == highspy.HighsModelStatus.kNotset:
            results.termination_condition = TerminationCondition.unknown
        elif status == highspy.HighsModelStatus.kLoadError:
            results.termination_condition = TerminationCondition.error
        elif status == highspy.HighsModelStatus.kModelError:
            results.termination_condition = TerminationCondition.error
        elif status == highspy.HighsModelStatus.kPresolveError:
            results.termination_condition = TerminationCondition.error
        elif status == highspy.HighsModelStatus.kSolveError:
            results.termination_condition = TerminationCondition.error
        elif status == highspy.HighsModelStatus.kPostsolveError:
            results.termination_condition = TerminationCondition.error
        elif status == highspy.HighsModelStatus.kModelEmpty:
            results.termination_condition = TerminationCondition.unknown
        elif status == highspy.HighsModelStatus.kOptimal:
            results.termination_condition = TerminationCondition.optimal
        elif status == highspy.HighsModelStatus.kInfeasible:
            results.termination_condition = TerminationCondition.infeasible
        elif status == highspy.HighsModelStatus.kUnboundedOrInfeasible:
            results.termination_condition = TerminationCondition.infeasibleOrUnbounded
        elif status == highspy.HighsModelStatus.kUnbounded:
            results.termination_condition = TerminationCondition.unbounded
        elif status == highspy.HighsModelStatus.kObjectiveBound:
            results.termination_condition = TerminationCondition.objectiveLimit
        elif status == highspy.HighsModelStatus.kObjectiveTarget:
            results.termination_condition = TerminationCondition.objectiveLimit
        elif status == highspy.HighsModelStatus.kTimeLimit:
            results.termination_condition = TerminationCondition.maxTimeLimit
        elif status == highspy.HighsModelStatus.kIterationLimit:
            results.termination_condition = TerminationCondition.maxIterations
        elif status == getattr(highspy.HighsModelStatus, "kSolutionLimit", NOTSET):
            # kSolutionLimit was introduced in HiGHS v1.5.3 for MIP-related limits
            results.termination_condition = TerminationCondition.maxIterations
        elif status == highspy.HighsModelStatus.kUnknown:
            results.termination_condition = TerminationCondition.unknown
        else:
            logger.warning(f'Received unhandled {status=} from solver HiGHS.')
            results.termination_condition = TerminationCondition.unknown

        timer.start('load solution')
        self._sol = highs.getSolution()
        has_feasible_solution = False
        if results.termination_condition == TerminationCondition.optimal:
            has_feasible_solution = True
        elif results.termination_condition in {
            TerminationCondition.objectiveLimit,
            TerminationCondition.maxIterations,
            TerminationCondition.maxTimeLimit,
        }:
            if self._sol.value_valid:
                has_feasible_solution = True

        if config.load_solution:
            if has_feasible_solution:
                if results.termination_condition != TerminationCondition.optimal:
                    logger.warning(
                        'Loading a feasible but suboptimal solution. '
                        'Please set load_solution=False and check '
                        'results.termination_condition and '
                        'results.found_feasible_solution() before loading a solution.'
                    )
                self.load_vars()
            else:
                raise RuntimeError(
                    'A feasible solution was not found, so no solution can be loaded. '
                    'If using the appsi.solvers.Highs interface, you can '
                    'set opt.config.load_solution=False. If using the environ.SolverFactory '
                    'interface, you can set opt.solve(model, load_solutions = False). '
                    'Then you can check results.termination_condition and '
                    'results.best_feasible_objective before loading a solution.'
                )
        timer.stop('load solution')

        info = highs.getInfo()
        results.best_objective_bound = None
        results.best_feasible_objective = None
        if self._objective is not None:
            if has_feasible_solution:
                results.best_feasible_objective = info.objective_function_value
            if info.mip_node_count == -1:
                if has_feasible_solution:
                    results.best_objective_bound = info.objective_function_value
                else:
                    results.best_objective_bound = None
            else:
                results.best_objective_bound = info.mip_dual_bound

        return results

    def load_vars(self, vars_to_load=None):
        for v, val in self.get_primals(vars_to_load=vars_to_load).items():
            v.set_value(val, skip_validation=True)
        StaleFlagManager.mark_all_as_stale(delayed=True)

    def get_primals(self, vars_to_load=None, solution_number=0):
        if self._sol is None or not self._sol.value_valid:
            raise RuntimeError(
                'Solver does not currently have a valid solution. Please '
                'check the termination condition.'
            )

        res = ComponentMap()
        if vars_to_load is None:
            var_ids_to_load = list()
            for v, ref_info in self._referenced_variables.items():
                using_cons, using_sos, using_obj = ref_info
                if using_cons or using_sos or (using_obj is not None):
                    var_ids_to_load.append(v)
        else:
            var_ids_to_load = [id(v) for v in vars_to_load]

        var_vals = self._sol.col_value

        for v_id in var_ids_to_load:
            v = self._vars[v_id][0]
            v_ndx = self._pyomo_var_to_solver_var_map[v_id]
            res[v] = var_vals[v_ndx]

        return res

    def get_reduced_costs(self, vars_to_load=None):
        if self._sol is None or not self._sol.dual_valid:
            raise RuntimeError(
                'Solver does not currently have valid reduced costs. Please '
                'check the termination condition.'
            )
        res = ComponentMap()
        if vars_to_load is None:
            var_ids_to_load = list(self._vars.keys())
        else:
            var_ids_to_load = [id(v) for v in vars_to_load]

        var_vals = self._sol.col_dual

        for v_id in var_ids_to_load:
            v = self._vars[v_id][0]
            v_ndx = self._pyomo_var_to_solver_var_map[v_id]
            res[v] = var_vals[v_ndx]

        return res

    def get_duals(self, cons_to_load=None):
        if self._sol is None or not self._sol.dual_valid:
            raise RuntimeError(
                'Solver does not currently have valid duals. Please '
                'check the termination condition.'
            )

        res = dict()
        if cons_to_load is None:
            cons_to_load = list(self._pyomo_con_to_solver_con_map.keys())

        duals = self._sol.row_dual

        for c in cons_to_load:
            c_ndx = self._pyomo_con_to_solver_con_map[c]
            res[c] = duals[c_ndx]

        return res

    def get_slacks(self, cons_to_load=None):
        if self._sol is None or not self._sol.value_valid:
            raise RuntimeError(
                'Solver does not currently have valid slacks. Please '
                'check the termination condition.'
            )

        res = dict()
        if cons_to_load is None:
            cons_to_load = list(self._pyomo_con_to_solver_con_map.keys())

        slacks = self._sol.row_value

        for c in cons_to_load:
            c_ndx = self._pyomo_con_to_solver_con_map[c]
            res[c] = slacks[c_ndx]

        return res



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
from pyomo.util.report_scaling import report_scaling
import logging
from pyomo.common import unittest
from pyomo.common.log import LoggingIntercept
from io import StringIO
import re


class TestReportScaling(unittest.TestCase):
    def test_report_scaling(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(list(range(5)))
        m.c = pyo.Constraint(list(range(4)))
        m.p = pyo.Param(initialize=0, mutable=True)

        m.x[0].setlb(-1)
        m.x[0].setub(1)
        m.x[1].setlb(1e5)
        m.x[1].setub(1e7)
        m.x[2].setlb(-1e5)
        m.x[2].setub(0)
        m.x[3].setlb(-20)
        m.x[3].setub(20)

        m.obj1 = pyo.Objective(expr=1e-8 * m.x[0] + pyo.exp(m.x[3]) + m.x[1] * m.x[2])
        m.obj2 = pyo.Objective(expr=m.x[0] * m.x[3] + m.x[1] ** 2)

        m.c[0] = m.x[0] + m.x[3] == 0
        m.c[1] = 1 / m.x[1] == 1
        m.c[2] = m.x[1] * m.x[3] == 1
        m.c[3] = m.x[3] + m.p * m.x[0] == 1

        out = StringIO()
        with LoggingIntercept(out, 'pyomo.util.report_scaling', level=logging.INFO):
            report_scaling(m)

        expected = """

The following variables are not bounded. Please add bounds.
          LB          UB    Var
        -inf         inf    x[4]

The following variables have large bounds. Please scale them.
          LB          UB    Var
    1.00e+05    1.00e+07    x[1]
   -1.00e+05    0.00e+00    x[2]

The following objectives have potentially large coefficients. Please scale them.
obj1
         Coef LB     Coef UB    Var
        2.06e-09    4.85e+08    x[3]
       -1.00e+05    0.00e+00    x[1]
        1.00e+05    1.00e+07    x[2]

obj2
         Coef LB     Coef UB    Var
        2.00e+05    2.00e+07    x[1]

The following objectives have small coefficients.
obj1
         Coef LB     Coef UB    Var
        1.00e-08    1.00e-08    x[0]

The following constraints have potentially large coefficients. Please scale them.
c[2]
         Coef LB     Coef UB    Var
        1.00e+05    1.00e+07    x[3]

The following constraints have small coefficients.
c[1]
         Coef LB     Coef UB    Var
       -1.00e-10   -1.00e-14    x[1]

The following constraints have bodies with large bounds. Please scale them.
          LB          UB    Constraint
   -2.00e+08    2.00e+08    c[2]

"""

        self.assertEqual(out.getvalue(), expected)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Tests for the MindtPy solver."""

import sys
import pyomo.common.unittest as unittest
from pyomo.contrib.mindtpy.tests.eight_process_problem import EightProcessFlowsheet
from pyomo.contrib.mindtpy.tests.MINLP_simple import SimpleMINLP as SimpleMINLP
from pyomo.contrib.mindtpy.tests.MINLP3_simple import SimpleMINLP as SimpleMINLP3
from pyomo.contrib.mindtpy.tests.constraint_qualification_example import (
    ConstraintQualificationExample,
)
from pyomo.environ import SolverFactory, value
from pyomo.opt import TerminationCondition

required_nlp_solvers = 'ipopt'
required_mip_solvers = ['cplex_persistent', 'gurobi_persistent']
available_mip_solvers = [
    s for s in required_mip_solvers if SolverFactory(s).available(False)
]

if (
    SolverFactory(required_nlp_solvers).available(exception_flag=False)
    and available_mip_solvers
):
    subsolvers_available = True
else:
    subsolvers_available = False


model_list = [
    EightProcessFlowsheet(convex=True),
    ConstraintQualificationExample(),
    SimpleMINLP(),
    SimpleMINLP3(),
]


def known_solver_failure(mip_solver, model):
    if (
        mip_solver == 'gurobi_persistent'
        and model.name in {'DuranEx3', 'SimpleMINLP'}
        and sys.platform.startswith('win')
        and SolverFactory(mip_solver).version()[:3] == (9, 5, 0)
    ):
        sys.stderr.write(
            f"Skipping sub-test {model.name} with {mip_solver} due to known "
            f"failure when running Gurobi 9.5.0 on Windows\n"
        )
        return True
    return False


@unittest.skipIf(
    not subsolvers_available,
    'Required subsolvers %s are not available'
    % ([required_nlp_solvers] + required_mip_solvers),
)
class TestMindtPy(unittest.TestCase):
    """Tests for the MindtPy solver plugin."""

    def check_optimal_solution(self, model, places=1):
        for var in model.optimal_solution:
            self.assertAlmostEqual(
                var.value, model.optimal_solution[var], places=places
            )

    @unittest.skipUnless(
        'cplex_persistent' in available_mip_solvers,
        'cplex_persistent solver is not available',
    )
    def test_LPNLP_CPLEX(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='OA',
                    mip_solver='cplex_persistent',
                    nlp_solver=required_nlp_solvers,
                    single_tree=True,
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=1
                )
                self.check_optimal_solution(model)

    @unittest.skipUnless(
        'gurobi_persistent' in available_mip_solvers,
        'gurobi_persistent solver is not available',
    )
    def test_LPNLP_Gurobi(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='OA',
                    mip_solver='gurobi_persistent',
                    nlp_solver=required_nlp_solvers,
                    single_tree=True,
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=1
                )
                self.check_optimal_solution(model)

    def test_RLPNLP_L1(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                for mip_solver in available_mip_solvers:
                    results = opt.solve(
                        model,
                        strategy='OA',
                        mip_solver=mip_solver,
                        nlp_solver=required_nlp_solvers,
                        single_tree=True,
                        add_regularization='level_L1',
                    )

                    self.assertIn(
                        results.solver.termination_condition,
                        [TerminationCondition.optimal, TerminationCondition.feasible],
                    )
                    self.assertAlmostEqual(
                        value(model.objective.expr), model.optimal_value, places=1
                    )
                    self.check_optimal_solution(model)

    def test_RLPNLP_L2(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                for mip_solver in available_mip_solvers:
                    if known_solver_failure(mip_solver, model):
                        continue
                    results = opt.solve(
                        model,
                        strategy='OA',
                        mip_solver=mip_solver,
                        nlp_solver=required_nlp_solvers,
                        single_tree=True,
                        add_regularization='level_L2',
                    )

                    self.assertIn(
                        results.solver.termination_condition,
                        [TerminationCondition.optimal, TerminationCondition.feasible],
                    )
                    self.assertAlmostEqual(
                        value(model.objective.expr), model.optimal_value, places=1
                    )
                    self.check_optimal_solution(model)

    def test_RLPNLP_Linf(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                for mip_solver in available_mip_solvers:
                    results = opt.solve(
                        model,
                        strategy='OA',
                        mip_solver=mip_solver,
                        nlp_solver=required_nlp_solvers,
                        single_tree=True,
                        add_regularization='level_L_infinity',
                    )

                    self.assertIn(
                        results.solver.termination_condition,
                        [TerminationCondition.optimal, TerminationCondition.feasible],
                    )
                    self.assertAlmostEqual(
                        value(model.objective.expr), model.optimal_value, places=1
                    )
                    self.check_optimal_solution(model)

    def test_RLPNLP_grad_lag(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                for mip_solver in available_mip_solvers:
                    results = opt.solve(
                        model,
                        strategy='OA',
                        mip_solver=mip_solver,
                        nlp_solver=required_nlp_solvers,
                        single_tree=True,
                        add_regularization='grad_lag',
                    )

                    self.assertIn(
                        results.solver.termination_condition,
                        [TerminationCondition.optimal, TerminationCondition.feasible],
                    )
                    self.assertAlmostEqual(
                        value(model.objective.expr), model.optimal_value, places=1
                    )
                    self.check_optimal_solution(model)

    def test_RLPNLP_hess_lag(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                for mip_solver in available_mip_solvers:
                    if known_solver_failure(mip_solver, model):
                        continue
                    results = opt.solve(
                        model,
                        strategy='OA',
                        mip_solver=mip_solver,
                        nlp_solver=required_nlp_solvers,
                        single_tree=True,
                        add_regularization='hess_lag',
                    )

                    self.assertIn(
                        results.solver.termination_condition,
                        [TerminationCondition.optimal, TerminationCondition.feasible],
                    )
                    self.assertAlmostEqual(
                        value(model.objective.expr), model.optimal_value, places=1
                    )
                    self.check_optimal_solution(model)

    def test_RLPNLP_hess_only_lag(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                for mip_solver in available_mip_solvers:
                    if known_solver_failure(mip_solver, model):
                        continue
                    results = opt.solve(
                        model,
                        strategy='OA',
                        mip_solver=mip_solver,
                        nlp_solver=required_nlp_solvers,
                        single_tree=True,
                        add_regularization='hess_only_lag',
                    )

                    self.assertIn(
                        results.solver.termination_condition,
                        [TerminationCondition.optimal, TerminationCondition.feasible],
                    )
                    self.assertAlmostEqual(
                        value(model.objective.expr), model.optimal_value, places=1
                    )
                    self.check_optimal_solution(model)

    def test_RLPNLP_sqp_lag(self):
        """Test the LP/NLP decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                for mip_solver in available_mip_solvers:
                    if known_solver_failure(mip_solver, model):
                        continue
                    results = opt.solve(
                        model,
                        strategy='OA',
                        mip_solver=mip_solver,
                        nlp_solver=required_nlp_solvers,
                        single_tree=True,
                        add_regularization='sqp_lag',
                    )

                    self.assertIn(
                        results.solver.termination_condition,
                        [TerminationCondition.optimal, TerminationCondition.feasible],
                    )
                    self.assertAlmostEqual(
                        value(model.objective.expr), model.optimal_value, places=1
                    )
                    self.check_optimal_solution(model)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import inspect
import pyomo.common.unittest as unittest

from pyomo.common import DeveloperError
from pyomo.core.base.disable_methods import disable_methods
from pyomo.common.modeling import NOTSET


class LocalClass:
    def __init__(self, name):
        self._name = name

    def __str__(self):
        return self._name


local_instance = LocalClass('local')


class _simple:
    def __init__(self, name):
        self.name = name
        self._d = 'd'
        self._e = 'e'

    def construct(self, data=None):
        return 'construct'

    def a(self):
        return 'a'

    def b(self):
        return 'b'

    def c(self):
        return 'c'

    @property
    def d(self):
        return self._d

    @d.setter
    def d(self, value='d'):
        self._d = value

    @property
    def e(self):
        return self._e

    @e.setter
    def e(self, value='e'):
        self._e = value

    def f(self, arg1, arg2=1, arg3=NOTSET, arg4=local_instance):
        return "f:%s,%s,%s,%s" % (arg1, arg2, arg3, arg4)

    @property
    def g(self):
        return 'g'

    @property
    def h(self):
        return 'h'


@disable_methods(
    (
        'a',
        ('b', 'custom_msg'),
        'd',
        ('e', 'custom_pmsg'),
        'f',
        'g',
        ('h', 'custom_pmsg'),
    )
)
class _abstract_simple(_simple):
    pass


class TestDisableMethods(unittest.TestCase):
    def test_signature(self):
        # check that signatures are properly preserved
        self.assertEqual(
            inspect.signature(_simple.construct),
            inspect.signature(_abstract_simple.construct),
        )
        self.assertEqual(
            inspect.signature(_simple.a), inspect.signature(_abstract_simple.a)
        )
        self.assertEqual(
            inspect.signature(_simple.b), inspect.signature(_abstract_simple.b)
        )
        self.assertEqual(
            inspect.signature(_simple.c), inspect.signature(_abstract_simple.c)
        )
        self.assertEqual(
            inspect.signature(_simple.d.fget),
            inspect.signature(_abstract_simple.d.fget),
        )
        self.assertEqual(
            inspect.signature(_simple.d.fset),
            inspect.signature(_abstract_simple.d.fset),
        )
        self.assertEqual(
            inspect.signature(_simple.e.fget),
            inspect.signature(_abstract_simple.e.fget),
        )
        self.assertEqual(
            inspect.signature(_simple.e.fset),
            inspect.signature(_abstract_simple.e.fset),
        )

        self.assertEqual(
            inspect.signature(_simple.f), inspect.signature(_abstract_simple.f)
        )

        self.assertEqual(
            inspect.signature(_simple.g.fget),
            inspect.signature(_abstract_simple.g.fget),
        )
        self.assertIsNone(_simple.g.fset)
        self.assertIsNone(_abstract_simple.g.fset)
        self.assertEqual(
            inspect.signature(_simple.h.fget),
            inspect.signature(_abstract_simple.h.fget),
        )
        self.assertIsNone(_simple.h.fset)
        self.assertIsNone(_abstract_simple.h.fset)

    def test_disable(self):
        x = _abstract_simple('foo')
        self.assertIs(type(x), _abstract_simple)
        self.assertIsInstance(x, _simple)
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot access 'a' on _abstract_simple "
            "'foo' before it has been constructed",
        ):
            x.a()
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot custom_msg _abstract_simple "
            "'foo' before it has been constructed",
        ):
            x.b()
        self.assertEqual(x.c(), 'c')
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot access property 'd' on _abstract_simple "
            "'foo' before it has been constructed",
        ):
            x.d
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot set property 'd' on _abstract_simple "
            "'foo' before it has been constructed",
        ):
            x.d = 1
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot custom_pmsg _abstract_simple "
            "'foo' before it has been constructed",
        ):
            x.e
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot custom_pmsg _abstract_simple "
            "'foo' before it has been constructed",
        ):
            x.e = 1

        # Verify that the wrapper function enforces the same API as the
        # wrapped function
        with self.assertRaisesRegex(TypeError, r"f\(\) takes "):
            x.f(1, 2, 3, 4, 5)
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot access 'f' on _abstract_simple "
            "'foo' before it has been constructed",
        ):
            x.f(1, 2)

        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot access property 'g' on _abstract_simple "
            "'foo' before it has been constructed",
        ):
            x.g
        with self.assertRaisesRegex(
            AttributeError, "(can't set attribute)|(object has no setter)"
        ):
            x.g = 1
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot custom_pmsg _abstract_simple "
            "'foo' before it has been constructed",
        ):
            x.h
        with self.assertRaisesRegex(
            AttributeError, "(can't set attribute)|(object has no setter)"
        ):
            x.h = 1

        self.assertEqual(x.construct(), 'construct')
        self.assertIs(type(x), _simple)
        self.assertIsInstance(x, _simple)
        self.assertEqual(x.a(), 'a')
        self.assertEqual(x.b(), 'b')
        self.assertEqual(x.c(), 'c')
        self.assertEqual(x.d, 'd')
        x.d = 1
        self.assertEqual(x.d, 1)
        self.assertEqual(x.e, 'e')
        x.e = 2
        self.assertEqual(x.e, 2)
        self.assertEqual(x.f(1, 2), 'f:1,2,NOTSET,local')
        self.assertEqual(x.g, 'g')
        self.assertEqual(x.h, 'h')

    def test_bad_api(self):
        with self.assertRaisesRegex(
            DeveloperError,
            r"Cannot disable method not_there on <class '.*\.foo'>",
            normalize_whitespace=True,
        ):

            @disable_methods(('a', 'not_there'))
            class foo(_simple):
                pass



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
from pyomo.core import (
    Var,
    Constraint,
    Objective,
    Set,
    minimize,
    exp,
    ConcreteModel,
    LogicalConstraint,
    exactly,
    lnot,
    lor,
    BooleanVar,
    land,
)
from pyomo.dae import Integral, DerivativeVar, ContinuousSet
from pyomo.gdp import Disjunct, Disjunction


def build_model(mode_transfer=False):
    model = ConcreteModel()

    # Set
    model.stage = Set(initialize=[1, 2, 3, 4])
    model.mode = Set(initialize=[1, 2, 3])

    model.t1 = ContinuousSet(bounds=(0, 1))
    model.t2 = ContinuousSet(bounds=(1, 2))
    model.t3 = ContinuousSet(bounds=(2, 3))
    model.t4 = ContinuousSet(bounds=(3, 4))

    # Variables
    model.x1 = Var(model.t1, bounds=(0, 10))
    model.x2 = Var(model.t2, bounds=(0, 10))
    model.x3 = Var(model.t3, bounds=(0, 10))
    model.x4 = Var(model.t4, bounds=(0, 10))
    model.u1 = Var(bounds=(-4, 4))
    model.u2 = Var(bounds=(-4, 4))
    model.u3 = Var(bounds=(-4, 4))
    model.u4 = Var(bounds=(-4, 4))

    # Dynamic model
    model.dxdt1 = DerivativeVar(model.x1, wrt=model.t1)
    model.dxdt2 = DerivativeVar(model.x2, wrt=model.t2)
    model.dxdt3 = DerivativeVar(model.x3, wrt=model.t3)
    model.dxdt4 = DerivativeVar(model.x4, wrt=model.t4)

    # logic constraint
    model.stage_mode = Disjunct(model.stage * model.mode)
    model.d = Disjunction(model.stage)

    # Stage 1

    def stage1_mode1_dynamic(disjunct, t):
        model = disjunct.model()
        return model.dxdt1[t] == -model.x1[t] * exp(model.x1[t] - 1) + model.u1

    model.stage_mode[1, 1].mode1_dynamic_constraint = Constraint(
        model.t1, rule=stage1_mode1_dynamic
    )

    def stage1_mode2_dynamic(disjunct, t):
        model = disjunct.model()
        return model.dxdt1[t] == (0.5 * model.x1[t] ** 3 + model.u1) / 20

    model.stage_mode[1, 2].mode2_dynamic_constraint = Constraint(
        model.t1, rule=stage1_mode2_dynamic
    )

    def stage1_mode3_dynamic(disjunct, t):
        model = disjunct.model()
        return model.dxdt1[t] == (model.x1[t] ** 2 + model.u1) / (t + 20)

    model.stage_mode[1, 3].mode3_dynamic_constraint = Constraint(
        model.t1, rule=stage1_mode3_dynamic
    )

    # Stage 2

    def stage2_mode1_dynamic(disjunct, t):
        model = disjunct.model()
        return model.dxdt2[t] == -model.x2[t] * exp(model.x2[t] - 1) + model.u2

    model.stage_mode[2, 1].mode1_dynamic_constraint = Constraint(
        model.t2, rule=stage2_mode1_dynamic
    )

    def stage2_mode2_dynamic(disjunct, t):
        model = disjunct.model()
        return model.dxdt2[t] == (0.5 * model.x2[t] ** 3 + model.u2) / 20

    model.stage_mode[2, 2].mode2_dynamic_constraint = Constraint(
        model.t2, rule=stage2_mode2_dynamic
    )

    def stage2_mode3_dynamic(disjunct, t):
        model = disjunct.model()
        return model.dxdt2[t] == (model.x2[t] ** 2 + model.u2) / (t + 20)

    model.stage_mode[2, 3].mode3_dynamic_constraint = Constraint(
        model.t2, rule=stage2_mode3_dynamic
    )

    # Stage 3

    def stage3_mode1_dynamic(disjunct, t):
        model = disjunct.model()
        return model.dxdt3[t] == -model.x3[t] * exp(model.x3[t] - 1) + model.u3

    model.stage_mode[3, 1].mode1_dynamic_constraint = Constraint(
        model.t3, rule=stage3_mode1_dynamic
    )

    def stage3_mode2_dynamic(disjunct, t):
        model = disjunct.model()
        return model.dxdt3[t] == (0.5 * model.x3[t] ** 3 + model.u3) / 20

    model.stage_mode[3, 2].mode2_dynamic_constraint = Constraint(
        model.t3, rule=stage3_mode2_dynamic
    )

    def stage3_mode3_dynamic(disjunct, t):
        model = disjunct.model()
        return model.dxdt3[t] == (model.x3[t] ** 2 + model.u3) / (t + 20)

    model.stage_mode[3, 3].mode3_dynamic_constraint = Constraint(
        model.t3, rule=stage3_mode3_dynamic
    )

    # Stage 4

    def stage4_mode1_dynamic(disjunct, t):
        model = disjunct.model()
        return model.dxdt4[t] == -model.x4[t] * exp(model.x4[t] - 1) + model.u4

    model.stage_mode[4, 1].mode1_dynamic_constraint = Constraint(
        model.t4, rule=stage4_mode1_dynamic
    )

    def stage4_mode2_dynamic(disjunct, t):
        model = disjunct.model()
        return model.dxdt4[t] == (0.5 * model.x4[t] ** 3 + model.u4) / 20

    model.stage_mode[4, 2].mode2_dynamic_constraint = Constraint(
        model.t4, rule=stage4_mode2_dynamic
    )

    def stage4_mode3_dynamic(disjunct, t):
        model = disjunct.model()
        return model.dxdt4[t] == (model.x4[t] ** 2 + model.u4) / (t + 20)

    model.stage_mode[4, 3].mode3_dynamic_constraint = Constraint(
        model.t4, rule=stage4_mode3_dynamic
    )

    model.d[1] = [
        model.stage_mode[1, 1],
        model.stage_mode[1, 2],
        model.stage_mode[1, 3],
    ]
    model.d[2] = [
        model.stage_mode[2, 1],
        model.stage_mode[2, 2],
        model.stage_mode[2, 3],
    ]
    model.d[3] = [
        model.stage_mode[3, 1],
        model.stage_mode[3, 2],
        model.stage_mode[3, 3],
    ]
    model.d[4] = [
        model.stage_mode[4, 1],
        model.stage_mode[4, 2],
        model.stage_mode[4, 3],
    ]

    if mode_transfer:
        model.lc1 = LogicalConstraint(
            expr=exactly(
                1,
                model.stage_mode[1, 1].indicator_var,
                model.stage_mode[1, 2].indicator_var,
                model.stage_mode[1, 3].indicator_var,
            )
        )
        model.lc2 = LogicalConstraint(
            expr=exactly(
                1,
                model.stage_mode[2, 1].indicator_var,
                model.stage_mode[2, 2].indicator_var,
                model.stage_mode[2, 3].indicator_var,
            )
        )
        model.lc3 = LogicalConstraint(
            expr=exactly(
                1,
                model.stage_mode[3, 1].indicator_var,
                model.stage_mode[3, 2].indicator_var,
                model.stage_mode[3, 3].indicator_var,
            )
        )
        model.lc4 = LogicalConstraint(
            expr=exactly(
                1,
                model.stage_mode[4, 1].indicator_var,
                model.stage_mode[4, 2].indicator_var,
                model.stage_mode[4, 3].indicator_var,
            )
        )
        model.transfer_stage1 = Set(initialize=[2, 3, 4])
        model.transfer_stage2 = Set(initialize=[2, 3, 4, 5])
        model.mode_stransfer_set = Set(initialize=[1, 2])
        model.mode_transfer = BooleanVar(
            model.transfer_stage2, model.mode_stransfer_set
        )
        model.mode_transfer_lc1 = LogicalConstraint(
            expr=exactly(
                1,
                model.mode_transfer[2, 1],
                model.mode_transfer[3, 1],
                model.mode_transfer[4, 1],
                model.mode_transfer[5, 1],
            )
        )
        model.mode_transfer_lc2 = LogicalConstraint(
            expr=exactly(
                1,
                model.mode_transfer[2, 2],
                model.mode_transfer[3, 2],
                model.mode_transfer[4, 2],
                model.mode_transfer[5, 2],
            )
        )

        def _mode_transfer_rule1(model, stage):
            return model.mode_transfer[stage, 1].equivalent_to(
                land(
                    model.stage_mode[stage - 1, 1].indicator_var,
                    model.stage_mode[stage, 2].indicator_var,
                )
            )

        model.mode_transfer2mode_choice_lc1 = LogicalConstraint(
            model.transfer_stage1, rule=_mode_transfer_rule1
        )

        def _mode_transfer_rule2(model, stage):
            return model.mode_transfer[stage, 2].equivalent_to(
                land(
                    model.stage_mode[stage - 1, 2].indicator_var,
                    model.stage_mode[stage, 3].indicator_var,
                )
            )

        model.mode_transfer2mode_choice_lc2 = LogicalConstraint(
            model.transfer_stage1, rule=_mode_transfer_rule2
        )

        def _mode_transfer_rule3(model, stage):
            return model.mode_transfer[stage, 2].implies(
                lor(
                    model.mode_transfer[stage1, 1]
                    for stage1 in model.transfer_stage1
                    if stage1 < stage
                )
            )

        model.mode_transfer2mode_choice_lc3 = LogicalConstraint(
            model.transfer_stage1, rule=_mode_transfer_rule3
        )

        def _mode_transfer_rule4(model):
            return model.mode_transfer[5, 1].implies(
                lnot(
                    lor(
                        model.stage_mode[stage1, 2].indicator_var
                        for stage1 in model.stage
                    )
                )
            )

        model.mode_transfer2mode_choice_lc4 = LogicalConstraint(
            rule=_mode_transfer_rule4
        )

        def _mode_transfer_rule5(model):
            return model.mode_transfer[5, 2].implies(
                lnot(
                    lor(
                        model.stage_mode[stage1, 3].indicator_var
                        for stage1 in model.stage
                    )
                )
            )

        model.mode_transfer2mode_choice_lc5 = LogicalConstraint(
            rule=_mode_transfer_rule5
        )

    # Sequence constraint
    def _sequence_rule1(model, stage):
        if stage == 1:
            return Constraint.Skip
        else:
            return model.stage_mode[stage, 2].indicator_var.implies(
                lor(
                    model.stage_mode[stage2, 1].indicator_var
                    for stage2 in model.stage
                    if stage2 < stage
                )
            )

    model.seq1 = LogicalConstraint(model.stage, rule=_sequence_rule1)
    model.stage_mode[1, 2].indicator_var.fix(False)

    def _sequence_rule2(model, stage):
        if stage == 4:
            return Constraint.Skip
        else:
            return model.stage_mode[stage, 2].indicator_var.implies(
                lnot(
                    lor(
                        model.stage_mode[stage2, 1].indicator_var
                        for stage2 in model.stage
                        if stage2 > stage
                    )
                )
            )

    model.seq2 = LogicalConstraint(model.stage, rule=_sequence_rule2)

    def _sequence_rule3(model, stage):
        if stage <= 1:
            return Constraint.Skip
        else:
            return model.stage_mode[stage, 3].indicator_var.implies(
                lor(
                    model.stage_mode[stage2, 2].indicator_var
                    for stage2 in model.stage
                    if stage2 < stage
                )
            )

    model.seq3 = LogicalConstraint(model.stage, rule=_sequence_rule3)
    model.stage_mode[1, 3].indicator_var.fix(False)
    model.stage_mode[2, 3].indicator_var.fix(False)

    def _sequence_rule4(model, stage):
        if stage == 4:
            return Constraint.Skip
        else:
            return model.stage_mode[stage, 3].indicator_var.implies(
                lnot(
                    lor(
                        model.stage_mode[stage2, 2].indicator_var
                        for stage2 in model.stage
                        if stage2 > stage
                    )
                )
            )

    model.seq4 = LogicalConstraint(model.stage, rule=_sequence_rule4)

    model.c1 = Constraint(expr=model.x1[0] == 1)
    model.c2 = Constraint(expr=model.x1[1] == model.x2[1])
    model.c3 = Constraint(expr=model.x2[2] == model.x3[2])
    model.c4 = Constraint(expr=model.x3[3] == model.x4[3])

    # Objective function
    model.intx1 = Integral(
        model.t1, wrt=model.t1, rule=lambda model, t: model.x1[t] ** 2
    )
    model.intx2 = Integral(
        model.t2, wrt=model.t2, rule=lambda model, t: model.x2[t] ** 2
    )
    model.intx3 = Integral(
        model.t3, wrt=model.t3, rule=lambda model, t: model.x3[t] ** 2
    )
    model.intx4 = Integral(
        model.t4, wrt=model.t4, rule=lambda model, t: model.x4[t] ** 2
    )

    model.obj = Objective(
        expr=-(model.intx1 + model.intx2 + model.intx3 + model.intx4), sense=minimize
    )
    return model



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the PyUtilib project
# Copyright (c) 2008 Sandia Corporation.
# This software is distributed under the BSD License.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
# ____________________________________________________________________________________

"""This module provides general utilities for working with the file system

.. autosummary::

   this_file
   this_file_dir
   find_path
   find_file
   find_dir
   find_library
   find_executable
   import_file
   PathManager
   PathData
"""

import glob
import inspect
import logging
import os
import platform
import importlib.util
import subprocess
import sys

from . import envvar
from .dependencies import ctypes
from .deprecation import deprecated, relocated_module_attribute

relocated_module_attribute('StreamIndenter', 'pyomo.common.formatting', version='6.2')


def this_file(stack_offset=1):
    """Returns the file name for the module that calls this function.

    This function is more reliable than __file__ on platforms like
    Windows and in situations where the program has called
    os.chdir().

    """
    # __file__ fails if script is called in different ways on Windows
    # __file__ fails if someone does os.chdir() before
    # sys.argv[0] also fails because it does not always contains the path
    callerFrame = inspect.currentframe()
    while stack_offset:
        callerFrame = callerFrame.f_back
        stack_offset -= 1
    frameName = callerFrame.f_code.co_filename
    if frameName and frameName[0] == '<' and frameName[-1] == '>':
        return frameName
    return os.path.abspath(inspect.getfile(callerFrame))


def this_file_dir(stack_offset=1):
    """Returns the directory containing the module that calls this function."""
    return os.path.dirname(this_file(stack_offset=1 + stack_offset))


PYOMO_ROOT_DIR = os.path.dirname(os.path.dirname(this_file_dir()))


def find_path(
    name,
    validate,
    cwd=True,
    mode=os.R_OK,
    ext=None,
    pathlist=[],
    allow_pathlist_deep_references=True,
):
    """Locate a path, given a set of search parameters

    Parameters
    ----------
    name : str
        The name to locate.  The name may contain references to a user's
        home directory (``~user``), environment variables
        (``${HOME}/bin``), and shell wildcards (``?`` and ``*``); all of
        which will be expanded.

    validate : function
        A function to call to validate the path (used by find_file and
        find_dir to discriminate files and directories)

    cwd : bool
        Start by looking in the current working directory
        [default: True]

    mode : mask
        If not None, only return entries that can be accessed for
        reading/writing/executing.  Valid values are the inclusive OR of
        {os.R_OK, os.W_OK, os.X_OK} [default: ``os.R_OK``]

    ext : str or iterable of str
        If not None, also look for name+ext [default: None]

    pathlist : str or iterable of str
        A list of strings containing paths to search, each string
        contains a single path.  If pathlist is a string, then it is
        first split using os.pathsep to generate the pathlist
        [default: ``[]``].

    allow_pathlist_deep_references : bool
       If allow_pathlist_deep_references is True and the name
       appears to be a relative path, allow deep reference matches
       relative to directories in the pathlist (e.g., if name is
       ``foo/my.exe`` and ``/usr/bin`` is in the pathlist, then
       :py:func:`find_file` could return ``/usr/bin/foo/my.exe``).  If
       allow_pathlist_deep_references is False and the name appears
       to be a relative path, then only matches relative to the current
       directory are allowed (assuming cwd==True).  [default: True]

    Notes
    -----
        find_path uses glob, so the path and/or name may contain
        wildcards.  The first matching entry is returned.

    """

    # Support shell-style paths like ~user and $HOME/bin
    name = os.path.expanduser(os.path.expandvars(name))

    locations = []
    if cwd:
        locations.append(os.getcwd())

    if allow_pathlist_deep_references or os.path.basename(name) == name:
        if isinstance(pathlist, str):
            locations.extend(pathlist.split(os.pathsep))
        else:
            locations.extend(pathlist)

    extlist = ['']
    if ext:
        if isinstance(ext, str):
            extlist.append(ext)
        else:
            extlist.extend(ext)

    for path in locations:
        if not path:
            continue
        for _ext in extlist:
            for test in glob.glob(os.path.join(path, name + _ext)):
                if not validate(test):
                    continue
                if mode is not None and not os.access(test, mode):
                    continue
                return os.path.abspath(test)
    return None


def find_file(
    filename,
    cwd=True,
    mode=os.R_OK,
    ext=None,
    pathlist=[],
    allow_pathlist_deep_references=True,
):
    """Locate a file, given a set of search parameters

    Parameters
    ----------
    filename : str

        The file name to locate.  The file name may contain references
        to a user's home directory (``~user``), environment variables
        (``${HOME}/bin``), and shell wildcards (``?`` and ``*``); all of
        which will be expanded.

    cwd : bool
        Start by looking in the current working directory
        [default: True]

    mode : mask
        If not None, only return files that can be accessed for
        reading/writing/executing.  Valid values are the inclusive OR of
        {os.R_OK, os.W_OK, os.X_OK} [default: ``os.R_OK``]

    ext : str or iterable of str
        If not None, also look for filename+ext [default: None]

    pathlist : str or iterable of str
        A list of strings containing paths to search, each string
        contains a single path.  If pathlist is a string, then it is
        first split using os.pathsep to generate the pathlist
        [default: ``[]``].

    allow_pathlist_deep_references : bool
       If allow_pathlist_deep_references is True and the filename
       appears to be a relative path, allow deep reference matches
       relative to directories in the pathlist (e.g., if filename is
       ``foo/my.exe`` and ``/usr/bin`` is in the pathlist, then
       :py:func:`find_file` could return ``/usr/bin/foo/my.exe``).  If
       allow_pathlist_deep_references is False and the filename appears
       to be a relative path, then only matches relative to the current
       directory are allowed (assuming cwd==True).  [default: True]

    Notes
    -----
        find_file uses glob, so the path and/or file name may contain
        wildcards.  The first matching file is returned.

    """
    return find_path(
        filename,
        os.path.isfile,
        cwd=cwd,
        mode=mode,
        ext=ext,
        pathlist=pathlist,
        allow_pathlist_deep_references=allow_pathlist_deep_references,
    )


def find_dir(
    dirname, cwd=True, mode=os.R_OK, pathlist=[], allow_pathlist_deep_references=True
):
    """Locate a directory, given a set of search parameters

    Parameters
    ----------
    dirname : str
        The directory name to locate.  The name may contain references
        to a user's home directory (``~user``), environment variables
        (``${HOME}/bin``), and shell wildcards (``?`` and ``*``); all of
        which will be expanded.

    cwd : bool
        Start by looking in the current working directory
        [default: True]

    mode : mask
        If not None, only return directories that can be accessed for
        reading/writing/executing.  Valid values are the inclusive OR of
        {os.R_OK, os.W_OK, os.X_OK} [default: ``os.R_OK``]

    pathlist : str or iterable of str
        A list of strings containing paths to search, each string
        contains a single path.  If pathlist is a string, then it is
        first split using os.pathsep to generate the pathlist
        [default: ``[]``].

    allow_pathlist_deep_references : bool
       If allow_pathlist_deep_references is True and the dirname
       appears to be a relative path, allow deep reference matches
       relative to directories in the pathlist (e.g., if dirname is
       ``foo/bar`` and ``/usr/bin`` is in the pathlist, then
       :py:func:`find_dir` could return ``/usr/bin/foo/bar``).  If
       allow_pathlist_deep_references is False and the dirname appears
       to be a relative path, then only matches relative to the current
       directory are allowed (assuming cwd==True).  [default: True]

    Notes
    -----
        find_dir uses glob, so the path and/or directory name may contain
        wildcards.  The first matching directory is returned.

    """
    return find_path(
        dirname,
        os.path.isdir,
        cwd=cwd,
        mode=mode,
        pathlist=pathlist,
        allow_pathlist_deep_references=allow_pathlist_deep_references,
    )


_exeExt = {
    'linux': None,
    'freebsd': None,
    'windows': '.exe',
    'cygwin': '.exe',
    'darwin': None,
}

_libExt = {
    'linux': ('.so', '.so.*'),
    'freebsd': ('.so', '.so.*'),
    'windows': ('.dll', '.pyd'),
    'cygwin': ('.dll', '.so', '.so.*'),
    'darwin': ('.dylib', '.so', '.so.*'),
}


def _system():
    system = platform.system().lower()
    for c in '.-_':
        system = system.split(c)[0]
    return system


def _path():
    return (os.environ.get('PATH', '') or os.defpath).split(os.pathsep)


def find_library(libname, cwd=True, include_PATH=True, pathlist=None):
    """Find a dynamic library using find_file to search typical locations.

    Finds a specified library (file) by searching a specified set of
    paths.  This routine will look for the specified file name, as well
    as looking for the filename followed by architecture-specific
    extensions (e.g., `.dll`, `.so`, or `.dynlib`).  Note that as this
    uses :py:func:find_file(), the filename and search paths may contain
    wildcards.

    If the explicit path search fails to locate a library, then this
    returns the result from passing the basename (with 'lib' and extension
    removed) to ctypes.util.find_library()

    Parameters
    ----------
    libname : str
        The library name to search for

    cwd : bool
        Start by looking in the current working directory
        [default: True]

    include_PATH : bool
        Include the executable search PATH at the end of the list of
        directories to search. [default: True]

    pathlist : str or list of str
        List of paths to search for the file.  If None, then pathlist
        will default to the local Pyomo configuration library directory
        (and the local Pyomo binary directory if include_PATH is set)
        and the contents of LD_LIBRARY_PATH.  If a string, then the
        string is split using os.pathsep.  [default: None]

    Notes
    -----
        find_library() uses :py:func:`find_file` with
        ``allow_pathlist_deep_references=True``, so libnames containing
        relative paths will be matched relative to all paths in
        pathlist.

    """
    if pathlist is None:
        # Note: PYOMO_CONFIG_DIR/lib comes before LD_LIBRARY_PATH, and
        # PYOMO_CONFIG_DIR/bin comes immediately before PATH
        pathlist = [os.path.join(envvar.PYOMO_CONFIG_DIR, 'lib')]
        pathlist.extend(os.environ.get('LD_LIBRARY_PATH', '').split(os.pathsep))
        if include_PATH:
            pathlist.append(os.path.join(envvar.PYOMO_CONFIG_DIR, 'bin'))
    elif isinstance(pathlist, str):
        pathlist = pathlist.split(os.pathsep)
    else:
        pathlist = list(pathlist)
    if include_PATH:
        pathlist.extend(_path())
    ext = _libExt.get(_system(), None)
    # Search 1: original filename (with extensions) in our paths
    lib = find_file(libname, cwd=cwd, ext=ext, pathlist=pathlist)
    if lib is None and not libname.startswith('lib'):
        # Search 2: prepend 'lib' (with extensions) in our paths
        lib = find_file('lib' + libname, cwd=cwd, ext=ext, pathlist=pathlist)
    if lib is not None:
        return lib
    # Search 3: use ctypes.util.find_library (which expects 'lib' and
    # extension to be removed from the name)
    libname_base, ext = os.path.splitext(os.path.basename(libname))
    if libname_base.startswith('lib') and _system() != 'windows':
        libname_base = libname_base[3:]
    if ext.lower().startswith(('.so', '.dll', '.dylib')):
        lib = ctypes.util.find_library(libname_base)
    else:
        lib = ctypes.util.find_library(libname)
    if lib and os.path.sep not in lib:
        # work around https://github.com/python/cpython/issues/65241,
        # where python does not return the absolute path on *nix
        try:
            libname = lib + ' '
            with subprocess.Popen(
                ['/sbin/ldconfig', '-p'],
                stdin=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                stdout=subprocess.PIPE,
                env={'LC_ALL': 'C', 'LANG': 'C'},
            ) as p:
                for line in os.fsdecode(p.stdout.read()).splitlines():
                    if line.lstrip().startswith(libname):
                        return os.path.realpath(line.split()[-1])
        except:
            pass
    return lib


def find_executable(exename, cwd=True, include_PATH=True, pathlist=None):
    """Find an executable using find_file to search typical locations.

    Finds a specified executable by searching a specified set of paths.
    This routine will look for the specified file name, as well as
    looking for the filename followed by architecture-specific
    extensions (e.g., ``.exe``).  Note that as this uses
    :py:func:`find_file()`, the filename and search paths may contain
    wildcards.

    Parameters
    ----------
    exename : str
        The executable file name to search for

    cwd : bool
        Start by looking in the current working directory
        [default: True]

    include_PATH : bool
        Include the executable search PATH at the end of the list of
        directories to search. [default: True]

    pathlist : str or list of str
        List of paths to search for the file.  If None, then pathlist
        will default to the local Pyomo configuration binary directory.
        If a string, then the string is split using os.pathsep.
        [Default: None]

    Notes
    -----
        find_executable() uses :py:func:`find_file` with
        ``allow_pathlist_deep_references=False``, so search strings
        containing relative paths will only be matched relative to the
        current working directory.  This prevents confusion in the case
        where a user called ``find_executable("./foo")`` and forgot to copy
        ``foo`` into the local directory, but this function picked up
        another ``foo`` in the user's PATH that they did not want to use.

    """
    if pathlist is None:
        pathlist = [os.path.join(envvar.PYOMO_CONFIG_DIR, 'bin')]
    elif isinstance(pathlist, str):
        pathlist = pathlist.split(os.pathsep)
    else:
        pathlist = list(pathlist)
    if include_PATH:
        pathlist.extend(_path())
    ext = _exeExt.get(_system(), None)
    return find_file(
        exename,
        cwd=cwd,
        ext=ext,
        mode=os.R_OK | os.X_OK,
        pathlist=pathlist,
        allow_pathlist_deep_references=False,
    )


def import_file(path, clear_cache=False, infer_package=True, module_name=None):
    """
    Import a module given the full path/filename of the file.
    Replaces import_file from pyutilib (Pyomo 6.0.0).

    This function returns the module object that is created.

    Parameters
    ----------
    path: str
        Full path to .py file.
    clear_cache: bool
        Remove module if already loaded. The default is False.
    """
    path = os.path.normpath(
        os.path.abspath(os.path.expanduser(os.path.expandvars(path)))
    )
    if not os.path.exists(path):
        raise FileNotFoundError('File does not exist. Check path.')
    module_dir, module_file = os.path.split(path)
    if module_name is None:
        module_name, module_ext = os.path.splitext(module_file)
    if infer_package:
        while module_dir and os.path.exists(os.path.join(module_dir, '__init__.py')):
            module_dir, mod = os.path.split(module_dir)
            module_name = mod + '.' + module_name
    if clear_cache and module_name in sys.modules:
        del sys.modules[module_name]
    sys.path.insert(0, module_dir)
    try:
        spec = importlib.util.spec_from_file_location(module_name, path)
        module = importlib.util.module_from_spec(spec)
        if module_name not in sys.modules:
            sys.modules[module_name] = module
        spec.loader.exec_module(module)
    finally:
        sys.path.pop(0)
    return module


def to_legal_filename(name, universal=False) -> str:
    """Convert a string to a legal filename on the current platform.

    This converts a candidate file name (not a path) and converts it to
    a legal file name on the current platform.  This includes replacing
    any unallowable characters (including the path separator) with
    underscores (``_``), and on some platforms, enforcing restrictions
    on the allowable final character.

    Parameters
    ----------
    name : str

        The original (desired) file name

    universal : bool

        If True, this will attempt a form of "universal" standardization
        that uses the most restrictive set of character translations and
        rules.  Currently, ``universal=True`` is equivalent to running
        the Windows translations.

    """
    if envvar.is_windows or universal:
        tr = getattr(to_legal_filename, 'tr', None)
        if tr is None:
            # Windows illegal characters: 0-31, plus < > : " / \ | ? *
            _illegal = r'<>:"/\|?*' + ''.join(map(chr, range(32)))
            tr = to_legal_filename.tr = str.maketrans(_illegal, '_' * len(_illegal))
        # Remove illegal characters
        name = name.translate(tr)
        if name:
            # Windows allows filenames to end with space or dot, but the
            # file explorer can't interact with them
            if name[-1] in ' .':
                name = name[:-1] + '_'
            # Similarly, starting with a space is generally a bad idea
            if name[0] == ' ':
                name = '_' + name[1:]
    else:
        name = name.replace('/', '_').replace(chr(0), '_')
    return name


class PathData:
    """An object for storing and managing a :py:class:`PathManager` path"""

    def __init__(self, manager, name):
        self._mngr = manager
        self._registered_name = name
        self._path = None
        self._path_override = None
        self._status = None

    def path(self):
        """Return the full, normalized path to the registered path entry.

        If the object is not found (or was marked "disabled"),
        ``path()`` returns None.

        """
        if self._status is None:
            if self._path_override:
                target = self._path_override
            else:
                target = self._registered_name
            tmp = self._mngr._find(target, pathlist=self._mngr.pathlist)
            self._path = tmp if tmp else self._path_override
            self._status = bool(tmp)
        return self._path

    def set_path(self, value):
        self._path_override = value
        self.rehash()
        if not self._status:
            logging.getLogger('pyomo.common').warning(
                "explicitly setting the path for '%s' to an "
                "invalid object or nonexistent location ('%s')"
                % (self._registered_name, value)
            )

    @deprecated(
        "get_path() is deprecated; use pyomo.common.Executable(name).path()",
        version='5.6.2',
    )
    def get_path(self):
        return self.path()

    def disable(self):
        """Disable this path entry

        This method "disables" this path entry by marking it as "not
        found".  Disabled entries return False for `available()` and
        None for `path()`.  The disabled status will persist until the
        next call to `rehash()`.

        """
        self._status = False
        self._path = None

    def available(self):
        """Returns True if the registered path is available.

        Entries are available if the object was found found in the
        search locations and has not been explicitly disabled.

        """
        if self._status is None:
            self.path()
        return self._status

    def rehash(self):
        """Requery the location of this path entry

        This method derives its name from the csh command of the same
        name, which rebuilds the hash table of executables reachable
        through the PATH.

        """
        self._status = None
        self.path()

    def __nonzero__(self):
        """Alias for ``available()``."""
        return self.available()

    # "if <obj>" triggers __nonzero__ in Py2 and __bool__ in Py3
    __bool__ = __nonzero__

    def __str__(self):
        ans = self.path()
        if not ans:
            return ""
        return ans


class ExecutableData(PathData):
    """A :py:class:`PathData` class specifically for executables."""

    @property
    def executable(self):
        """Get (or set) the path to the executable"""
        return self.path()

    @executable.setter
    def executable(self, value):
        self.set_path(value)


class PathManager:
    """The PathManager defines a registry class for path locations

    The :py:class:`PathManager` defines a class very similar to the
    :py:class:`CachedFactory` class; however it does not register type
    constructors.  Instead, it registers instances of
    :py:class:`PathData` (or :py:class:`ExecutableData`).  These
    contain the resolved path to the directory object under which the
    :py:class:`PathData` object was registered.  We do not use
    the PyUtilib ``register_executable`` and ``registered_executable``
    functions so that we can automatically include Pyomo-specific
    locations in the search path (namely the ``PYOMO_CONFIG_DIR``).

    Users will generally interact with this class through global
    instances of this class (``pyomo.common.Executable`` and
    ``pyomo.common.Library``).

    Users are not required or expected to register file names with the
    :py:class:`PathManager`; they will be automatically registered
    upon first use.  Generally, users interact through the ``path()``
    and ``available()`` methods:

    .. doctest::
        :hide:

        >>> import pyomo.common
        >>> import os
        >>> from stat import S_IXUSR, S_IXGRP, S_IXOTH
        >>> _testfile = os.path.join(
        ...    pyomo.common.envvar.PYOMO_CONFIG_DIR, 'bin', 'demo_exec_file')
        >>> _del_testfile = not os.path.exists(_testfile)
        >>> if _del_testfile:
        ...     open(_testfile,'w').close()
        ...     _mode = os.stat(_testfile).st_mode
        ...     os.chmod(_testfile, _mode | S_IXUSR | S_IXGRP | S_IXOTH)

    .. doctest::

        >>> from pyomo.common import Executable
        >>> if Executable('demo_exec_file').available():
        ...     loc = Executable('demo_exec_file').path()
        ...     print(os.path.isfile(loc))
        True
        >>> print(os.access(loc, os.X_OK))
        True

    For convenience, :py:meth:`available()` and :py:meth:`path()` are
    available by casting the :py:class:`PathData` object requrned
    from ``Executable`` or ``Library`` to either a ``bool`` or ``str``:

    .. doctest::

        >>> if Executable('demo_exec_file'):
        ...     cmd = "%s --help" % Executable('demo_exec_file')

    The :py:class:`PathManager` caches the location / existence of
    the target directory entry.  If something in the environment changes
    (e.g., the PATH) or the file is created or removed after the first
    time a client queried the location or availability, the
    PathManager will return incorrect information.  You can cause
    the :py:class:`PathManager` to refresh its cache by calling
    ``rehash()`` on either the :py:class:`PathData` (for the
    single file) or the :py:class:`PathManager` to refresh the
    cache for all files:

    .. doctest::

        >>> # refresh the cache for a single file
        >>> Executable('demo_exec_file').rehash()
        >>> # or all registered files
        >>> Executable.rehash()

    The ``Executable`` singleton looks for executables in the system
    ``PATH`` and in the list of directories specified by the ``pathlist``
    attribute.  ``Executable.pathlist`` defaults to a list containing the
    ``os.path.join(pyomo.common.envvar.PYOMO_CONFIG_DIR, 'bin')``.

    The ``Library`` singleton looks for executables in the system
    ``LD_LIBRARY_PATH``, ``PATH`` and in the list of directories
    specified by the ``pathlist`` attribute.  ``Library.pathlist``
    defaults to a list containing the
    ``os.path.join(pyomo.common.envvar.PYOMO_CONFIG_DIR, 'lib')``.

    Users may also override the normal file resolution by explicitly
    setting the location using :py:meth:`set_path`:

    .. doctest::

        >>> Executable('demo_exec_file').set_path(os.path.join(
        ...     pyomo.common.envvar.PYOMO_CONFIG_DIR, 'bin', 'demo_exec_file'))

    Explicitly setting the path is an absolute operation and will
    set the location whether or not that location points to an actual
    file.  Additionally, the explicit location will persist
    through calls to ``rehash()``.  If you wish to remove the explicit
    executable location, call ``set_path(None)``:

    .. doctest::

        >>> Executable('demo_exec_file').set_path(None)

    The ``Executable`` singleton uses :py:class:`ExecutableData`, an
    extended form of the :py:class:`PathData` class, which provides the
    ``executable`` property as an alias for :py:meth:`path()` and
    :py:meth:`set_path()`:

    .. doctest::

        >>> loc = Executable('demo_exec_file').executable
        >>> print(os.path.isfile(loc))
        True
        >>> Executable('demo_exec_file').executable = os.path.join(
        ...     pyomo.common.envvar.PYOMO_CONFIG_DIR, 'bin', 'demo_exec_file')
        >>> Executable('demo_exec_file').executable = None

    .. doctest::
        :hide:

        >>> if _del_testfile:
        ...     os.remove(_testfile)

    """

    def __init__(self, finder, dataClass):
        self._pathTo = {}
        self._find = finder
        self._dataClass = dataClass
        self.pathlist = None

    def __call__(self, path):
        if path not in self._pathTo:
            if isinstance(path, self._dataClass):
                return path
            self._pathTo[path] = self._dataClass(self, path)
        return self._pathTo[path]

    def rehash(self):
        """Requery the location of all registered paths

        This method derives its name from the csh command of the same
        name, which rebuilds the hash table of executables reachable
        through the PATH.

        """
        for _path in self._pathTo.values():
            _path.rehash()


#
# Define singleton objects for Pyomo / Users to interact with
#
class Executable:
    """Singleton executable registry

    This class cannot be instantiated.  Instead, calling this type will
    perform lookups in the underlying singleton :class:`PathManager`
    object and return instances of :class:`ExecutableData`.

    """

    _manager = PathManager(find_executable, ExecutableData)

    def __new__(cls, path) -> ExecutableData:
        return cls._manager(path)

    @classmethod
    def rehash(cls):
        """Requery the location of all registered executable paths

        This method derives its name from the csh command of the same
        name, which rebuilds the hash table of executables reachable
        through the PATH.

        """
        return cls._manager.rehash()


class Library:
    """Singleton library registry

    This class cannot be instantiated.  Instead, calling this type will
    perform lookups in the underlying singleton :class:`PathManager`
    object and return instances of :class:`PathData`.

    """

    _manager = PathManager(find_library, PathData)

    def __new__(cls, path) -> PathData:
        return cls._manager(path)

    @classmethod
    def rehash(cls):
        """Requery the location of all registered library paths

        This method derives its name from the csh command of the same
        name, which rebuilds the hash table of executables reachable
        through the PATH.

        """
        return cls._manager.rehash()


@deprecated(
    "pyomo.common.register_executable(name) has been deprecated; "
    "explicit registration is no longer necessary",
    version='5.6.2',
)
def register_executable(name, validate=None):
    # Setting to None will cause Executable to re-search the pathlist
    return Executable(name).rehash()


@deprecated(
    """pyomo.common.registered_executable(name) has been deprecated; use
    pyomo.common.Executable(name).path() to get the path or
    pyomo.common.Executable(name).available() to get a bool indicating
    file availability.  Equivalent results can be obtained by casting
    Executable(name) to string or bool.""",
    version='5.6.2',
)
def registered_executable(name):
    ans = Executable(name)
    if ans.path() is None:
        return None
    else:
        return ans


@deprecated(
    "pyomo.common.unregister_executable(name) has been deprecated; "
    "use Executable(name).disable()",
    version='5.6.2',
)
def unregister_executable(name):
    Executable(name).disable()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Test the canonical expressions
#

import os
from filecmp import cmp
from io import StringIO

import pyomo.common.unittest as unittest
from pyomo.common.collections import OrderedSet
from pyomo.common.fileutils import this_file_dir
import pyomo.core.expr as EXPR
from pyomo.core.base import SymbolMap

from pyomo.environ import (
    ConcreteModel,
    Var,
    Param,
    Constraint,
    Objective,
    Block,
    sin,
    maximize,
    Binary,
    Suffix,
)
from pyomo.repn.plugins.baron_writer import expression_to_string

thisdir = this_file_dir()


class Test(unittest.TestCase):
    def _cleanup(self, fname):
        try:
            os.remove(fname)
        except OSError:
            pass

    def _get_fnames(self):
        class_name, test_name = self.id().split('.')[-2:]
        prefix = os.path.join(thisdir, test_name.replace("test_", "", 1))
        return prefix + ".bar.baseline", prefix + ".bar.out"

    def _check_baseline(self, model, **kwds):
        baseline_fname, test_fname = self._get_fnames()
        self._cleanup(test_fname)
        io_options = {"symbolic_solver_labels": True}
        io_options.update(kwds)
        model.write(test_fname, format="bar", io_options=io_options)
        try:
            self.assertTrue(cmp(test_fname, baseline_fname))
        except:
            with open(baseline_fname, 'r') as f1, open(test_fname, 'r') as f2:
                f1_contents = f1.read().replace(' ;', ';').split()
                f2_contents = f2.read().replace(' ;', ';').split()
                self.assertEqual(
                    f1_contents,
                    f2_contents,
                    "\n\nbaseline: %s\ntestFile: %s\n" % (baseline_fname, test_fname),
                )
        self._cleanup(test_fname)

    def _gen_expression(self, terms):
        terms = list(terms)
        expr = 0.0
        for term in terms:
            if type(term) is tuple:
                prodterms = list(term)
                prodexpr = 1.0
                for x in prodterms:
                    prodexpr *= x
                expr += prodexpr
            else:
                expr += term
        return expr

    def test_no_column_ordering_quadratic(self):
        model = ConcreteModel()
        model.a = Var()
        model.b = Var()
        model.c = Var()

        terms = [
            model.a,
            model.b,
            model.c,
            (model.a, model.a),
            (model.b, model.b),
            (model.c, model.c),
            (model.a, model.b),
            (model.a, model.c),
            (model.b, model.c),
        ]
        model.obj = Objective(expr=self._gen_expression(terms))
        model.con = Constraint(expr=self._gen_expression(terms) <= 1)
        self._check_baseline(model)

    def test_no_column_ordering_linear(self):
        model = ConcreteModel()
        model.a = Var()
        model.b = Var()
        model.c = Var()

        terms = [model.a, model.b, model.c]
        model.obj = Objective(expr=self._gen_expression(terms))
        model.con = Constraint(expr=self._gen_expression(terms) <= 1)
        self._check_baseline(model)

    def test_no_row_ordering(self):
        model = ConcreteModel()
        model.a = Var()

        components = {}
        components["obj"] = Objective(expr=model.a)
        components["con1"] = Constraint(expr=model.a >= 0)
        components["con2"] = Constraint(expr=model.a <= 1)
        components["con3"] = Constraint(expr=(0, model.a, 1))
        components["con4"] = Constraint([1, 2], rule=lambda m, i: model.a == i)

        for key in components:
            model.add_component(key, components[key])

        self._check_baseline(model, file_determinism=2)

    def test_var_on_other_model(self):
        other = ConcreteModel()
        other.a = Var()

        model = ConcreteModel()
        model.x = Var()
        model.c = Constraint(expr=other.a + 2 * model.x <= 0)
        model.obj = Objective(expr=model.x)
        self._check_baseline(model)

    def test_var_on_deactivated_block(self):
        model = ConcreteModel()
        model.x = Var()
        model.other = Block()
        model.other.a = Var()
        model.other.deactivate()
        model.c = Constraint(expr=model.other.a + 2 * model.x <= 0)
        model.obj = Objective(expr=model.x)
        self._check_baseline(model)

    def test_var_on_nonblock(self):
        class Foo(Block().__class__):
            def __init__(self, *args, **kwds):
                kwds.setdefault('ctype', Foo)
                super(Foo, self).__init__(*args, **kwds)

        model = ConcreteModel()
        model.x = Var()
        model.other = Foo()
        model.other.deactivate()
        model.other.a = Var()
        model.c = Constraint(expr=model.other.a + 2 * model.x <= 0)
        model.obj = Objective(expr=model.x)
        self._check_baseline(model)

    def test_trig_generates_exception(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 2 * 3.1415))
        m.obj = Objective(expr=sin(m.x))
        with self.assertRaisesRegex(
            RuntimeError,
            'The BARON .BAR format does not support the unary function "sin"',
        ):
            test_fname = self._get_fnames()[1]
            self._cleanup(test_fname)
            m.write(test_fname, format="bar")
        self._cleanup(test_fname)

    def test_exponential_NPV(self):
        m = ConcreteModel()
        m.x = Var()
        m.obj = Objective(expr=m.x**2)
        m.p = Param(initialize=1, mutable=True)
        m.c = Constraint(expr=m.x * m.p**1.2 == 0)
        self._check_baseline(m)

    def test_branching_priorities(self):
        m = ConcreteModel()
        m.x = Var(within=Binary)
        m.y = Var([1, 2, 3], within=Binary)
        m.c = Constraint(expr=m.y[1] * m.y[2] - 2 * m.x >= 0)
        m.obj = Objective(expr=m.y[1] + m.y[2], sense=maximize)
        m.priority = Suffix(direction=Suffix.EXPORT)
        m.priority[m.x] = 1
        # Note this checks that y[3] is filtered out
        m.priority[m.y] = 2
        self._check_baseline(m)

    def test_invalid_suffix(self):
        m = ConcreteModel()
        m.x = Var(within=Binary)
        m.y = Var([1, 2, 3], within=Binary)
        m.c = Constraint(expr=m.y[1] * m.y[2] - 2 * m.x >= 0)
        m.obj = Objective(expr=m.y[1] + m.y[2], sense=maximize)
        m.priorities = Suffix(direction=Suffix.EXPORT)
        m.priorities[m.x] = 1
        m.priorities[m.y] = 2
        with self.assertRaisesRegex(
            ValueError,
            "The BARON writer can not export suffix "
            "with name 'priorities'. Either remove it from "
            "the model or deactivate it.",
        ):
            m.write(StringIO(), format='bar')
        m._name = 'TestModel'
        with self.assertRaisesRegex(
            ValueError,
            "The BARON writer can not export suffix "
            "with name 'priorities'. Either remove it from "
            "the model 'TestModel' or deactivate it.",
        ):
            m.write(StringIO(), format='bar')
        p = m.priorities
        del m.priorities
        m.blk = Block()
        m.blk.sub = Block()
        m.blk.sub.priorities = p
        with self.assertRaisesRegex(
            ValueError,
            "The BARON writer can not export suffix "
            "with name 'priorities'. Either remove it from "
            "the block 'blk.sub' or deactivate it.",
        ):
            m.write(StringIO(), format='bar')


class TestToBaronVisitor(unittest.TestCase):
    def test_pow(self):
        variables = OrderedSet()
        smap = SymbolMap()

        m = ConcreteModel()
        m.x = Var(initialize=1)
        m.y = Var(initialize=2)
        m.p = Param(mutable=True, initialize=0)

        e = m.x**m.y
        test = expression_to_string(e, variables, smap)
        self.assertEqual(test, "exp((x) * log(y))")

        e = m.x ** (3 + EXPR.ProductExpression((m.p, m.y)))
        test = expression_to_string(e, variables, smap)
        self.assertEqual(test, "x ^ 3")

        e = (3 + EXPR.ProductExpression((m.p, m.y))) ** m.x
        test = expression_to_string(e, variables, smap)
        self.assertEqual(test, "3 ^ x")

    def test_issue_2819(self):
        m = ConcreteModel()
        m.x = Var()
        m.z = Var()
        t = 0.55
        m.x.fix(3.5)
        e = (m.x - 4) ** 2 + (m.z - 1) ** 2 - t

        variables = OrderedSet()
        smap = SymbolMap()
        test = expression_to_string(e, variables, smap)
        self.assertEqual(test, '(-0.5) ^ 2 + (z - 1) ^ 2 + (-0.55)')


# class TestBaron_writer(unittest.TestCase):
class XTestBaron_writer:
    """These tests verified that the BARON writer complained loudly for
    variables that were not on the model, not on an active block, or not
    on a Block ctype.  As we are relaxing that requirement throughout
    Pyomo, these tests have been disabled."""

    def _cleanup(self, fname):
        try:
            os.remove(fname)
        except OSError:
            pass

    def _get_fnames(self):
        class_name, test_name = self.id().split('.')[-2:]
        prefix = os.path.join(thisdir, test_name.replace("test_", "", 1))
        return prefix + ".bar.baseline", prefix + ".bar.out"

    def test_var_on_other_model(self):
        other = ConcreteModel()
        other.a = Var()

        model = ConcreteModel()
        model.x = Var()
        model.c = Constraint(expr=other.a + 2 * model.x <= 0)
        model.obj = Objective(expr=model.x)

        baseline_fname, test_fname = self._get_fnames()
        self._cleanup(test_fname)
        self.assertRaises(KeyError, model.write, test_fname, format='bar')
        self._cleanup(test_fname)

    def test_var_on_deactivated_block(self):
        model = ConcreteModel()
        model.x = Var()
        model.other = Block()
        model.other.a = Var()
        model.other.deactivate()
        model.c = Constraint(expr=model.other.a + 2 * model.x <= 0)
        model.obj = Objective(expr=model.x)

        baseline_fname, test_fname = self._get_fnames()
        self._cleanup(test_fname)
        self.assertRaises(KeyError, model.write, test_fname, format='bar')
        self._cleanup(test_fname)

    def test_var_on_nonblock(self):
        class Foo(Block().__class__):
            def __init__(self, *args, **kwds):
                kwds.setdefault('ctype', Foo)
                super(Foo, self).__init__(*args, **kwds)

        model = ConcreteModel()
        model.x = Var()
        model.other = Foo()
        model.other.a = Var()
        model.c = Constraint(expr=model.other.a + 2 * model.x <= 0)
        model.obj = Objective(expr=model.x)

        baseline_fname, test_fname = self._get_fnames()
        self._cleanup(test_fname)
        self.assertRaises(KeyError, model.write, test_fname, format='bar')
        self._cleanup(test_fname)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

logger = logging.getLogger(__name__)

from pyomo.common.dependencies import attempt_import

gurobipy, gurobi_available = attempt_import("gurobipy")

import pyomo.environ as pyo
import pyomo.common.errors
from pyomo.contrib.alternative_solutions import aos_utils, shifted_lp, solution
from pyomo.contrib import appsi


class NoGoodCutGenerator:
    def __init__(
        self,
        model,
        variable_groups,
        zero_threshold,
        orig_model,
        all_variables,
        orig_objective,
        num_solutions,
    ):
        self.model = model
        self.zero_threshold = zero_threshold
        self.variable_groups = variable_groups
        self.variables = aos_utils.get_model_variables(model)
        self.orig_model = orig_model
        self.all_variables = all_variables
        self.orig_objective = orig_objective
        self.solutions = []
        self.num_solutions = num_solutions

    def cut_generator_callback(self, cb_m, cb_opt, cb_where):
        if cb_where == gurobipy.GRB.Callback.MIPSOL:
            cb_opt.cbGetSolution(vars=self.variables)
            logger.info("***FOUND SOLUTION***")

            for var, index in self.model.var_map.items():
                var.set_value(var.lb + self.model.var_lower[index].value)
            sol = solution.Solution(
                self.orig_model, self.all_variables, objective=self.orig_objective
            )
            self.solutions.append(sol)

            if len(self.solutions) >= self.num_solutions:
                cb_opt._solver_model.terminate()
            num_non_zero = 0
            non_zero_basic_expr = 1
            for idx in range(len(self.variable_groups)):
                continuous_var, binary_var = self.variable_groups[idx]
                for var in continuous_var:
                    if continuous_var[var].value > self.zero_threshold:
                        num_non_zero += 1
                        non_zero_basic_expr += binary_var[var]
            # TODO: JLG - If we want to add the mixed binary case, I think we
            # need to do it here. Essentially we would want to continue to
            # build up the num_non_zero as follows
            # for binary in binary_vars:
            # if binary.value > 0.5:
            # num_non_zero += 1 - binary
            # else:
            # num_non_zero += binary
            new_con = self.model.cl.add(non_zero_basic_expr <= num_non_zero)
            cb_opt.cbLazy(new_con)


def enumerate_linear_solutions_soln_pool(
    model,
    num_solutions=10,
    rel_opt_gap=None,
    abs_opt_gap=None,
    zero_threshold=1e-5,
    solver_options={},
    tee=False,
):
    """
    Finds alternative optimal solutions for a (mixed-binary) linear program
    using Gurobi's solution pool feature.

    Parameters
    ----------
    model : ConcreteModel
        A concrete Pyomo model
    num_solutions : int
        The maximum number of solutions to generate.
    variables: None or a collection of Pyomo _GeneralVarData variables
        The variables for which bounds will be generated. None indicates
        that all variables will be included. Alternatively, a collection of
        _GeneralVarData variables can be provided.
    rel_opt_gap : float or None
        The relative optimality gap for the original objective for which
        variable bounds will be found. None indicates that a relative gap
        constraint will not be added to the model.
    abs_opt_gap : float or None
        The absolute optimality gap for the original objective for which
        variable bounds will be found. None indicates that an absolute gap
        constraint will not be added to the model.
    zero_threshold: float
        The threshold for which a continuous variables' value is considered
        to be equal to zero.
    solver_options : dict
        Solver option-value pairs to be passed to the solver.
    tee : boolean
        Boolean indicating that the solver output should be displayed.

    Returns
    -------
    solutions
        A list of Solution objects.
        [Solution]
    """
    logger.info("STARTING LP ENUMERATION ANALYSIS USING GUROBI SOLUTION POOL")
    #
    # Setup gurobi
    #
    if not gurobi_available:
        raise pyomo.common.errors.ApplicationError(f"Solver (gurobi) not available")

    all_variables = aos_utils.get_model_variables(model)
    for var in all_variables:
        if var.is_integer():
            raise pyomo.common.errors.ApplicationError(
                f"The enumerate_linear_solutions_soln_pool() function cannot be used with models that contain discrete variables"
            )

    opt = pyo.SolverFactory("gurobi")
    if not opt.available(exception_flag=False):
        raise ValueError(solver + " is not available")
    for parameter, value in solver_options.items():
        opt.options[parameter] = value

    logger.info("Performing initial solve of model.")
    results = opt.solve(model, tee=tee)
    status = results.solver.status
    condition = results.solver.termination_condition
    if condition != pyo.TerminationCondition.optimal:
        raise Exception(
            (
                "Model could not be solve. LP enumeration analysis "
                "cannot be applied, SolverStatus = {}, "
                "TerminationCondition = {}"
            ).format(status.value, condition.value)
        )

    orig_objective = aos_utils.get_active_objective(model)
    orig_objective_value = pyo.value(orig_objective)
    logger.info("Found optimal solution, value = {}.".format(orig_objective_value))

    aos_block = aos_utils._add_aos_block(model, name="_lp_enum")
    logger.info("Added block {} to the model.".format(aos_block))
    aos_utils._add_objective_constraint(
        aos_block, orig_objective, orig_objective_value, rel_opt_gap, abs_opt_gap
    )

    canonical_block = shifted_lp.get_shifted_linear_model(model)
    cb = canonical_block
    lower_index = list(cb.var_lower.keys())
    upper_index = list(cb.var_upper.keys())

    # w variables
    cb.basic_lower = pyo.Var(lower_index, domain=pyo.Binary)
    cb.basic_upper = pyo.Var(upper_index, domain=pyo.Binary)
    cb.basic_slack = pyo.Var(cb.slack_index, domain=pyo.Binary)

    # w upper bounds constraints
    def bound_lower_rule(m, var_index):
        return (
            m.var_lower[var_index]
            <= m.var_lower[var_index].ub * m.basic_lower[var_index]
        )

    cb.bound_lower = pyo.Constraint(lower_index, rule=bound_lower_rule)

    def bound_upper_rule(m, var_index):
        return (
            m.var_upper[var_index]
            <= m.var_upper[var_index].ub * m.basic_upper[var_index]
        )

    cb.bound_upper = pyo.Constraint(upper_index, rule=bound_upper_rule)

    def bound_slack_rule(m, var_index):
        return (
            m.slack_vars[var_index]
            <= m.slack_vars[var_index].ub * m.basic_slack[var_index]
        )

    cb.bound_slack = pyo.Constraint(cb.slack_index, rule=bound_slack_rule)

    cb.cl = pyo.ConstraintList()

    # TODO: If we go the mixed binary route we also want to list the binary variables
    variable_groups = [
        (cb.var_lower, cb.basic_lower),
        (cb.var_upper, cb.basic_upper),
        (cb.slack_vars, cb.basic_slack),
    ]
    cut_generator = NoGoodCutGenerator(
        cb,
        variable_groups,
        zero_threshold,
        model,
        all_variables,
        orig_objective,
        num_solutions,
    )

    opt = appsi.solvers.Gurobi()
    for parameter, value in solver_options.items():
        opt.gurobi_options[parameter] = value
    opt.config.stream_solver = True
    opt.config.load_solution = False
    opt.gurobi_options["LazyConstraints"] = 1
    opt.set_instance(cb)
    opt.set_callback(cut_generator.cut_generator_callback)
    opt.solve(cb)

    aos_block.deactivate()
    logger.info("COMPLETED LP ENUMERATION ANALYSIS")

    return cut_generator.solutions



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import numpy as np
from pyomo.common.dependencies import mpi4py
from pyomo.contrib.pynumero.sparse.mpi_block_vector import MPIBlockVector
from pyomo.contrib.pynumero.sparse.mpi_block_matrix import MPIBlockMatrix
from scipy.sparse import random


def main():
    comm = mpi4py.MPI.COMM_WORLD
    rank = comm.Get_rank()

    owners = [0, 1, 2, -1]
    x = MPIBlockVector(4, rank_owner=owners, mpi_comm=comm)

    owners = np.array([[0, -1, -1, 0], [-1, 1, -1, 1], [-1, -1, 2, 2]])
    a = MPIBlockMatrix(3, 4, rank_ownership=owners, mpi_comm=comm)

    np.random.seed(0)
    x.set_block(3, np.random.uniform(-10, 10, size=10))

    np.random.seed(rank)
    x.set_block(rank, np.random.uniform(-10, 10, size=10))
    a.set_block(rank, rank, random(10, 10, density=0.1))
    a.set_block(rank, 3, random(10, 10, density=0.1))

    b = a * x  # parallel matrix-vector dot product

    # check the answer
    local_x = x.make_local_copy().flatten()
    local_a = a.to_local_array()
    local_b = b.make_local_copy().flatten()

    err = np.abs(local_a.dot(local_x) - local_b).max()

    if rank == 0:
        print('error: ', err)

    return err


if __name__ == '__main__':
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
from operator import attrgetter

from pyomo.common.config import (
    ConfigBlock,
    ConfigValue,
    InEnum,
    document_class_CONFIG,
    In,
    ListOf,
)
from pyomo.common.gc_manager import PauseGC
from pyomo.common.timing import TicTocTimer
from pyomo.core.base import (
    Block,
    Objective,
    Constraint,
    Var,
    Param,
    Expression,
    SOSConstraint,
    Suffix,
    SymbolMap,
    minimize,
    ShortNameLabeler,
)
from pyomo.core.base.label import NumericLabeler
from pyomo.opt import WriterFactory
from pyomo.repn.linear import LinearRepnVisitor
from pyomo.repn.util import (
    FileDeterminism,
    FileDeterminism_to_SortComponents,
    OrderedVarRecorder,
    categorize_valid_components,
    ordered_active_constraints,
)
from pyomo.core.base import Set, RangeSet, ExternalFunction
from pyomo.network import Port

logger = logging.getLogger(__name__)
inf = float('inf')
neg_inf = float('-inf')


class GAMSWriterInfo(object):
    """Return type for GAMSWriter.write()

    Attributes
    ----------
    symbol_map: SymbolMap

        The :py:class:`SymbolMap` bimap between row/column labels and
        Pyomo components.
    """

    def __init__(self, var_symbol_map, con_symbol_map):
        self.var_symbol_map = var_symbol_map
        self.con_symbol_map = con_symbol_map


@WriterFactory.register(
    'gams_writer_v2', 'Generate the corresponding gms file (version 2).'
)
@document_class_CONFIG(methods=['write'])
class GAMSWriter(object):
    CONFIG = ConfigBlock('gamswriter')
    CONFIG.declare(
        'warmstart',
        ConfigValue(
            default=True,
            domain=bool,
            description="Warmstart by initializing model's variables to their values.",
        ),
    )
    CONFIG.declare(
        'symbolic_solver_labels',
        ConfigValue(
            default=False,
            domain=bool,
            description='Write variables/constraints using model names',
            doc="""
            Export variables and constraints to the gms file using human-readable
            text names derived from the corresponding Pyomo component names.
            """,
        ),
    )
    CONFIG.declare(
        'labeler',
        ConfigValue(
            default=None,
            description='Callable to use to generate symbol names in gms file',
        ),
    )
    CONFIG.declare(
        'solver',
        ConfigValue(
            default=None,
            domain=str,
            description='If None, GAMS will use default solver for model type.',
        ),
    )
    CONFIG.declare(
        'mtype',
        ConfigValue(
            default=None,
            domain=str,
            description='Model type',
            doc="""
            If None, will chose from lp, mip. nlp and minlp will be implemented
            in a future release.
            """,
        ),
    )
    CONFIG.declare(
        'gams_commands',
        ConfigValue(
            default=[],
            domain=ListOf(str),
            doc="""
            List of additional GAMS commands to write directly
            into the GAMS model file before the solve statement.
            Specifically for solvers.
            """,
        ),
    )
    CONFIG.declare(
        'skip_trivial_constraints',
        ConfigValue(
            default=False,
            domain=bool,
            description='Skip writing constraints whose body is constant',
        ),
    )
    CONFIG.declare(
        'output_fixed_variables',
        ConfigValue(
            default=False,
            domain=bool,
            description="""
            If True, output fixed variables as variables; otherwise,
            output numeric value
            """,
        ),
    )
    CONFIG.declare(
        'file_determinism',
        ConfigValue(
            default=FileDeterminism.ORDERED,
            domain=InEnum(FileDeterminism),
            description='How much effort to ensure file is deterministic',
            doc="""
            How much effort do we want to put into ensuring the
            GAMS file is written deterministically for a Pyomo model:

               - NONE (0) : None
               - ORDERED (10): rely on underlying component ordering (default)
               - SORT_INDICES (20) : sort keys of indexed components
               - SORT_SYMBOLS (30) : sort keys AND sort names (not declaration order)

            """,
        ),
    )
    CONFIG.declare(
        'put_results',
        ConfigValue(
            default='results',
            domain=str,
            doc="""
            Filename for optionally writing solution values and
            marginals.  If put_results_format is 'gdx', then GAMS
            will write solution values and marginals to
            GAMS_MODEL_p.gdx and solver statuses to
            {put_results}_s.gdx.  If put_results_format is 'dat',
            then solution values and marginals are written to
            (put_results).dat, and solver statuses to (put_results +
            'stat').dat.
            """,
        ),
    )
    CONFIG.declare(
        'put_results_format',
        ConfigValue(
            domain=In(["gdx", "dat"]),
            doc="""
            Format used for put_results, one of 'gdx', 'dat'.
            If not set, the format will default to 'gdx' if the gdx/gdxcc
            python module is available and 'dat' otherwise.
            """,
        ),
    )
    # NOTE: Taken from the lp_writer
    CONFIG.declare(
        'row_order',
        ConfigValue(
            default=None,
            description='Preferred constraint ordering',
            doc="""
            To use with ordered_active_constraints function.""",
        ),
    )

    def __init__(self):
        #: Instance configuration;
        #: see :ref:`pyomo.repn.plugins.gams_writer_v2.GAMSWriter::CONFIG`.
        self.config = self.CONFIG()

    def __call__(self, model, filename, solver_capability, io_options):
        if filename is None:
            filename = 'GAMS_MODEL' + ".gms"

        with open(filename, 'w', newline='') as FILE:
            info = self.write(model, FILE, **io_options)

        return filename, info.symbol_map

    def write(self, model, ostream, **options) -> GAMSWriterInfo:
        """Write a model in GMS format.

        Returns
        -------
        GAMSWriterInfo

        Parameters
        -------
        model: ConcreteModel
            The concrete Pyomo model to write out.

        ostream: io.TextIOBase
            The text output stream where the GMS "file" will be written.
            Could be an opened file or a io.StringIO.
        """
        config = options.pop('config', self.config)(options)

        # Pause the GC, as the walker that generates the compiled GMS
        # representation generates (and disposes of) a large number of
        # small objects.

        # NOTE: First pass write the model but needs variables/equations
        # definition first
        with PauseGC():
            return _GMSWriter_impl(ostream, config).write(model)


class _GMSWriter_impl(object):
    def __init__(self, ostream, config):
        # taken from lp_writer.py
        self.ostream = ostream
        self.config = config
        self.symbol_map = None

        # Taken from nl_writer.py
        self.symbolic_solver_labels = config.symbolic_solver_labels

        self.subexpression_cache = {}
        self.subexpression_order = None  # set to [] later
        self.external_functions = {}
        self.used_named_expressions = set()
        self.var_map = {}
        self.var_id_to_nl_map = {}
        self.next_V_line_id = 0
        self.pause_gc = None

    def write(self, model):
        timing_logger = logging.getLogger('pyomo.common.timing.writer')
        timer = TicTocTimer(logger=timing_logger)
        with_debug_timing = (
            timing_logger.isEnabledFor(logging.DEBUG) and timing_logger.hasHandlers()
        )

        # Caching some frequently-used objects into the locals()
        model_name = "GAMS_MODEL"
        symbolic_solver_labels = self.symbolic_solver_labels
        ostream = self.ostream
        config = self.config
        labeler = config.labeler
        var_labeler, con_labeler = None, None
        warmstart = config.warmstart

        sorter = FileDeterminism_to_SortComponents(config.file_determinism)
        if not config.put_results_format:
            config.put_results_format = 'gdx'

        component_map, unknown = categorize_valid_components(
            model,
            active=True,
            sort=sorter,
            valid={
                Block,
                Constraint,
                Var,
                Param,
                Expression,
                ExternalFunction,
                Set,
                RangeSet,
                Port,
            },
            targets={Suffix, SOSConstraint, Objective},
        )
        if unknown:
            raise ValueError(
                "The model ('%s') contains the following active components "
                "that the gams writer does not know how to process:\n\t%s"
                % (
                    model.name,
                    "\n\t".join(
                        "%s:\n\t\t%s" % (k, "\n\t\t".join(map(attrgetter('name'), v)))
                        for k, v in unknown.items()
                    ),
                )
            )

        if symbolic_solver_labels and (labeler is not None):
            raise ValueError(
                "GAMS writer: Using both the "
                "'symbolic_solver_labels' and 'labeler' "
                "I/O options is forbidden"
            )

        if symbolic_solver_labels:
            # Note that the Var and Constraint labelers must use the
            # same labeler, so that we can correctly detect name
            # collisions (which can arise when we truncate the labels to
            # the max allowable length.  GAMS requires all identifiers
            # to start with a letter.  We will (randomly) choose "s_"
            # (for 'shortened')
            var_labeler = con_labeler = ShortNameLabeler(
                60,
                prefix='s_',
                suffix='_',
                caseInsensitive=True,
                legalRegex='^[a-zA-Z]',
            )
        elif labeler is None:
            var_labeler = NumericLabeler('x')
            con_labeler = NumericLabeler('c')
        else:
            var_labeler = con_labeler = labeler

        self.var_symbol_map = SymbolMap(var_labeler)
        self.con_symbol_map = SymbolMap(con_labeler)
        self.var_order = {_id: i for i, _id in enumerate(self.var_map)}
        self.var_recorder = OrderedVarRecorder(self.var_map, self.var_order, sorter)

        visitor = LinearRepnVisitor(
            self.subexpression_cache, var_recorder=self.var_recorder
        )

        #
        # Tabulate constraints
        #
        skip_trivial_constraints = config.skip_trivial_constraints
        last_parent = None
        # NOTE: con_list Save the constraint representation and write it
        # after variables/equations declare
        con_list = {}
        for con in ordered_active_constraints(model, config):
            if with_debug_timing and con.parent_component() is not last_parent:
                timer.toc('Constraint %s', last_parent, level=logging.DEBUG)
                last_parent = con.parent_component()
            # Note: Constraint.to_bounded_expression(evaluate_bounds=True)
            # guarantees a return value that is either a (finite)
            # native_numeric_type, or None
            lb, body, ub = con.to_bounded_expression(True)

            repn = visitor.walk_expression(body)
            if repn.nonlinear is not None:
                raise ValueError(
                    f"Model constraint ({con.name}) contains nonlinear terms "
                    "which are currently not supported in the new gams_writer"
                )

            # Pull out the constant: we will move it to the bounds
            offset = repn.constant
            repn.constant = 0

            if repn.linear or getattr(repn, 'quadratic', None):
                pass
            else:
                if (
                    skip_trivial_constraints
                    and (lb is None or lb <= offset)
                    and (ub is None or ub >= offset)
                ):
                    continue

            con_symbol = con_labeler(con)
            declaration, definition, bounds = None, None, None
            if lb is not None:
                if ub is None:
                    label = f'{con_symbol}_lo'
                    self.con_symbol_map.addSymbol(con, label)
                    declaration = f'\n{label}.. '
                    definition = self.write_expression(ostream, repn)
                    bounds = f' =G= {(lb - offset)!s};'
                    con_list[label] = declaration + definition + bounds
                elif lb == ub:
                    label = f'{con_symbol}'
                    self.con_symbol_map.addSymbol(con, label)
                    declaration = f'\n{label}.. '
                    definition = self.write_expression(ostream, repn)
                    bounds = f' =E= {(lb - offset)!s};'
                    con_list[label] = declaration + definition + bounds
                else:
                    # We will need the constraint body twice.
                    # Procedure is taken from lp_writer.py
                    label = f'{con_symbol}_lo'
                    self.con_symbol_map.addSymbol(con, label)
                    declaration = f'\n{label}.. '
                    definition = self.write_expression(ostream, repn)
                    bounds = f' =G= {(lb - offset)!s};'
                    con_list[label] = declaration + definition + bounds
                    #
                    label = f'{con_symbol}_hi'
                    self.con_symbol_map.alias(con, label)
                    declaration = f'\n{label}.. '
                    definition = self.write_expression(ostream, repn)
                    bounds = f' =L= {(ub - offset)!s};'
                    con_list[label] = declaration + definition + bounds
            elif ub is not None:
                label = f'{con_symbol}_hi'
                self.con_symbol_map.addSymbol(con, label)
                declaration = f'\n{label}.. '
                definition = self.write_expression(ostream, repn)
                bounds = f' =L= {(ub - offset)!s};'
                con_list[label] = declaration + definition + bounds

        #
        # Process objective
        #
        if not component_map[Objective]:
            objectives = [Objective(expr=1)]
            objectives[0].construct()
        else:
            objectives = []
            for blk in component_map[Objective]:
                objectives.extend(
                    blk.component_data_objects(
                        Objective, active=True, descend_into=False, sort=sorter
                    )
                )
        if len(objectives) > 1:
            raise ValueError(
                "More than one active objective defined for input model '%s'; "
                "Cannot write legal gms file\nObjectives: %s"
                % (model.name, ' '.join(obj.name for obj in objectives))
            )

        obj = objectives[0]
        repn = visitor.walk_expression(obj.expr)
        if repn.nonlinear is not None:
            raise ValueError(
                f"Model objective ({obj.name}) contains nonlinear terms that "
                "is currently not supported in this new GAMSWriter"
            )

        label = self.con_symbol_map.getSymbol(obj, con_labeler)
        declaration = f'\n{label}.. -GAMS_OBJECTIVE '
        definition = self.write_expression(ostream, repn, True)
        bounds = f' =E= {(-repn.constant)!s};\n\n'
        con_list[label] = declaration + definition + bounds

        # Write the GAMS model
        ostream.write("$offlisting\n")
        # $offdigit ignores extra precise digits instead of erroring
        ostream.write("$offdigit\n\n")

        #
        # Write out variable declaration
        #
        integer_vars = []
        binary_vars = []
        var_bounds = {}
        getSymbolByObjectID = self.var_symbol_map.byObject.get

        ostream.write("VARIABLES \n")
        for vid, v in self.var_map.items():
            v_symbol = getSymbolByObjectID(vid, None)
            if not v_symbol:
                continue
            if v.is_continuous():
                ostream.write(f"\t{v_symbol} \n")
                lb, ub = v.bounds
                var_bounds[v_symbol] = (lb, ub)
            elif v.is_binary():
                binary_vars.append(v_symbol)
            elif v.is_integer():
                lb, ub = v.bounds
                var_bounds[v_symbol] = (lb, ub)
                integer_vars.append(v_symbol)

        ostream.write("\tGAMS_OBJECTIVE;\n\n")

        if integer_vars:
            ostream.write("\nINTEGER VARIABLES\n\t")
            ostream.write("\n\t".join(integer_vars) + ';\n\n')

        if binary_vars:
            ostream.write("\nBINARY VARIABLES\n\t")
            ostream.write("\n\t".join(binary_vars) + ';\n\n')

        #
        # Writing out the equations/constraints
        #
        ostream.write("EQUATIONS \n")
        for sym, con in con_list.items():
            ostream.write(f"\t{sym}\n")
        if con_list:
            ostream.write(";\n\n")

        for _, con in con_list.items():
            ostream.write(con)

        #
        # Handling variable bounds
        #
        warn_int_bounds = False
        for v, (lb, ub) in var_bounds.items():
            pyomo_v = self.var_symbol_map.bySymbol[v]
            if lb is None:
                lb = float("-inf")
            ostream.write(f'{v}.lo = {lb};\n')
            if ub is None:
                ub = float("inf")
            ostream.write(f'{v}.up = {ub};\n')
            if warmstart and pyomo_v.value is not None:
                ostream.write(f"{v}.l = {pyomo_v.value};\n")

        ostream.write(f'\nModel {model_name} / all /;\n')
        ostream.write(f'{model_name}.limrow = 0;\n')
        ostream.write(f'{model_name}.limcol = 0;\n')

        # CHECK FOR mtype flag based on variable domains - reals, integer
        if config.mtype is None:
            if binary_vars or integer_vars:
                config.mtype = 'mip'  # expand this to nlp, minlp
            else:
                config.mtype = 'lp'

        if config.put_results_format == 'gdx':
            ostream.write("option savepoint=1;\n")

        ostream.write("\n* START USER ADDITIONAL OPTIONS\n")
        if config.gams_commands:
            ostream.write('\n'.join(config.gams_commands))
        ostream.write("\n* END USER ADDITIONAL OPTIONS\n\n")

        ostream.write(
            "SOLVE %s USING %s %simizing GAMS_OBJECTIVE;\n"
            % (model_name, config.mtype, 'min' if obj.sense == minimize else 'max')
        )
        # Set variables to store certain statuses and attributes
        stat_vars = [
            'MODELSTAT',
            'SOLVESTAT',
            'OBJEST',
            'OBJVAL',
            'NUMVAR',
            'NUMEQU',
            'NUMDVAR',
            'NUMNZ',
            'ETSOLVE',
        ]
        ostream.write("\nScalars MODELSTAT 'model status', SOLVESTAT 'solve status';\n")
        ostream.write("MODELSTAT = %s.modelstat;\n" % model_name)
        ostream.write("SOLVESTAT = %s.solvestat;\n\n" % model_name)

        ostream.write("Scalar OBJEST 'best objective', OBJVAL 'objective value';\n")
        ostream.write("OBJEST = %s.objest;\n" % model_name)
        ostream.write("OBJVAL = %s.objval;\n\n" % model_name)

        ostream.write("Scalar NUMVAR 'number of variables';\n")
        ostream.write("NUMVAR = %s.numvar\n\n" % model_name)

        ostream.write("Scalar NUMEQU 'number of equations';\n")
        ostream.write("NUMEQU = %s.numequ\n\n" % model_name)

        ostream.write("Scalar NUMDVAR 'number of discrete variables';\n")
        ostream.write("NUMDVAR = %s.numdvar\n\n" % model_name)

        ostream.write("Scalar NUMNZ 'number of nonzeros';\n")
        ostream.write("NUMNZ = %s.numnz\n\n" % model_name)

        ostream.write("Scalar ETSOLVE 'time to execute solve statement';\n")
        ostream.write("ETSOLVE = %s.etsolve\n\n" % model_name)

        if config.put_results is not None:
            if config.put_results_format == 'gdx':
                ostream.write("\nexecute_unload '%s_s.gdx'" % config.put_results)
                for stat in stat_vars:
                    ostream.write(", %s" % stat)
                ostream.write(";\n")
            else:
                results = config.put_results + '.dat'
                ostream.write("\nfile results /'%s'/;" % results)
                ostream.write("\nresults.nd=15;")
                ostream.write("\nresults.nw=21;")
                ostream.write("\nput results;")
                ostream.write("\nput 'SYMBOL  :  LEVEL  :  MARGINAL' /;")
                for sym, var in self.var_symbol_map.bySymbol.items():
                    if var.parent_component().ctype is Var:
                        ostream.write("\nput %s ' ' %s.l ' ' %s.m /;" % (sym, sym, sym))
                for con in self.con_symbol_map.bySymbol.keys():
                    ostream.write("\nput %s ' ' %s.l ' ' %s.m /;" % (con, con, con))
                for con in self.con_symbol_map.aliases.keys():
                    ostream.write("\nput %s ' ' %s.l ' ' %s.m /;" % (con, con, con))
                ostream.write(
                    "\nput GAMS_OBJECTIVE ' ' GAMS_OBJECTIVE.l "
                    "' ' GAMS_OBJECTIVE.m;\n"
                )

                statresults = config.put_results + 'stat.dat'
                ostream.write("\nfile statresults /'%s'/;" % statresults)
                ostream.write("\nstatresults.nd=15;")
                ostream.write("\nstatresults.nw=21;")
                ostream.write("\nput statresults;")
                ostream.write("\nput 'SYMBOL   :   VALUE' /;")
                for stat in stat_vars:
                    ostream.write("\nput '%s' ' ' %s /;\n" % (stat, stat))

        timer.toc("Finished writing .gms file", level=logging.DEBUG)

        info = GAMSWriterInfo(self.var_symbol_map, self.con_symbol_map)
        return info

    def write_expression(self, ostream, expr, is_objective=False):
        if not is_objective:
            assert not expr.constant
        getSymbol = self.var_symbol_map.getSymbol
        getVarOrder = self.var_order.__getitem__
        getVar = self.var_map.__getitem__
        expr_str = ''
        if expr.linear:
            for vid, coef in sorted(
                expr.linear.items(), key=lambda x: getVarOrder(x[0])
            ):
                if coef < 0:
                    expr_str += f'{coef!s}*{getSymbol(getVar(vid))} \n'
                else:
                    expr_str += f'+ {coef!s} * {getSymbol(getVar(vid))} \n'

        return expr_str
